import numpy as np
import pandas as pd
from pandas import testing as tm
import pytest

from pandas_datareader import data as web

pytestmark = pytest.mark.stable


class TestEcondb(object):
    def test_infer_start_end_from_symbols(self):
        df = web.DataReader(
            (
                "dataset=NAMQ_10_GDP&v=Geopolitical entity (reporting)"
                "&h=TIME&from=2010-02-01&to=2018-10-01&GEO=[AL,AT,BE,BA,"
                "BG,HR,CY,CZ,DK,EE,EA19,FI,FR,DE,EL,HU,IS,IE,IT,XK,LV,LT,"
                "LU,MT,ME,NL,MK,NO,PL,PT,RO,RS,SK,SI,ES,SE,CH,TR,UK]"
                "&NA_ITEM=[B1GQ]&S_ADJ=[SCA]&UNIT=[CLV10_MNAC]"
            ),
            "econdb",
        )
        assert df.index[0].year == 2010
        assert df.index[-1].year == 2018

    @pytest.mark.xfail(reason="Dataset does not exist on Econdb")
    def test_get_cdh_e_fos(self):
        # EUROSTAT
        # Employed doctorate holders in non managerial and non professional
        # occupations by fields of science (%)
        df = web.DataReader(
            "dataset=CDH_E_FOS&GEO=NO,PL,PT,RU&FOS07=FOS1&Y_GRAD=TOTAL",
            "econdb",
            start=pd.Timestamp("2005-01-01"),
            end=pd.Timestamp("2010-01-01"),
        )
        assert isinstance(df, pd.DataFrame)
        assert df.shape == (2, 4)

        # the levels and not returned consistently for econdb
        names = list(df.columns.names)
        levels = [lvl.values.tolist() for lvl in list(df.columns.levels)]

        exp_col = pd.MultiIndex.from_product(levels, names=names)
        exp_idx = pd.DatetimeIndex(["2006-01-01", "2009-01-01"], name="TIME_PERIOD")

        values = np.array([[25.49, np.nan, 39.05, np.nan], [20.38, 25.1, 27.77, 38.1]])
        expected = pd.DataFrame(values, index=exp_idx, columns=exp_col)
        tm.assert_frame_equal(df, expected)

    def test_get_tourism(self):
        # OECD
        # TOURISM_INBOUND

        df = web.DataReader(
            "dataset=OE_TOURISM_INBOUND&COUNTRY=JPN,USA&VARIABLE=INB_ARRIVALS_TOTAL",
            "econdb",
            start=pd.Timestamp("2008-01-01"),
            end=pd.Timestamp("2012-01-01"),
        )
        df = df.astype(float)
        jp = np.array([8351000, 6790000, 8611000, 6219000, 8368000], dtype=float)
        us = np.array(
            [175702304, 160507424, 164079728, 167600272, 171320416], dtype=float
        )
        index = pd.date_range("2008-01-01", "2012-01-01", freq="AS", name="TIME_PERIOD")

        # check the values coming back are equal
        np.testing.assert_array_equal(df.values[:, 0], jp)
        np.testing.assert_array_equal(df.values[:, 1], us)

        # sometimes the country and variable columns are swapped
        df = df.swaplevel(2, 1, axis=1)
        for label, values in [("Japan", jp), ("United States", us)]:
            expected = pd.Series(
                values, index=index, name="Total international arrivals"
            )
            expected.index.freq = None
            tm.assert_series_equal(
                df[label]["Tourism demand surveys"]["Total international arrivals"],
                expected,
            )

    def test_bls(self):
        # BLS
        # CPI
        df = web.DataReader(
            "ticker=BLS_CU.CUSR0000SA0.M.US",
            "econdb",
            start=pd.Timestamp("2010-01-01"),
            end=pd.Timestamp("2013-01-27"),
        )

        assert df.loc["2010-05-01"][0] == 217.3

    def test_australia_gdp(self):
        df = web.DataReader(
            "dataset=ABS_GDP&to=2019-09-01&from=1959-09-01&h=TIME&v=Indicator", "econdb"
        )
        assert (
            df.loc[
                "2017-10-01",
                (
                    "GDP per capita: Current prices - National Accounts",
                    "Seasonally Adjusted",
                    "AUD",
                ),
            ]
            == 18329
        )



import datetime as dt

import pytest

from pandas_datareader import tsp as tsp

pytestmark = pytest.mark.stable


class TestTSPFunds(object):
    def test_get_allfunds(self):
        tspdata = tsp.TSPReader(start="2015-11-2", end="2015-11-2").read()

        assert len(tspdata) == 1

        assert round(tspdata["I Fund"][dt.datetime(2015, 11, 2)], 5) == 25.0058

    def test_get_one_fund(self):
        tspdata = tsp.TSPReader(
            start="2015-11-2", end="2015-11-2", symbols=("I Fund",)
        ).read()

        assert len(tspdata) == 1

        assert tspdata.columns.values.tolist() == ["I Fund"]

    def test_sanitize_response(self):
        class response(object):
            pass

        r = response()
        r.text = " , "
        ret = tsp.TSPReader._sanitize_response(r)
        assert ret == ""
        r.text = " a,b "
        ret = tsp.TSPReader._sanitize_response(r)
        assert ret == "a,b"



import pytest

from pandas_datareader import data as web
from pandas_datareader.exceptions import ImmediateDeprecationError

pytestmark = [
    pytest.mark.alpha_vantage,
]


class TestAVQuotes(object):
    @classmethod
    def setup_class(cls):
        pytest.importorskip("lxml")

    def test_invalid_symbol(self):
        with pytest.raises(ImmediateDeprecationError):
            web.get_quote_av("BADSYMBOL")

    def test_bad_multiple_symbol(self):
        with pytest.raises(ImmediateDeprecationError):
            web.get_quote_av(["AAPL", "BADSYMBOL"])

    def test_single_symbol(self):
        with pytest.raises(ImmediateDeprecationError):
            web.get_quote_av("AAPL", retry_count=6, pause=20.5)

    def test_multi_symbol(self):
        with pytest.raises(ImmediateDeprecationError):
            web.get_quote_av(["AAPL", "TSLA"], retry_count=6, pause=20.5)

    @pytest.mark.xfail(reason="May return NaN outside of market hours")
    def test_return_types(self):
        with pytest.raises(ImmediateDeprecationError):
            web.get_quote_av("AAPL", retry_count=6, pause=20.5)



import json
from urllib.parse import urlencode

import pandas as pd

from pandas_datareader.base import _BaseReader

# Data provided for free by IEX
# Data is furnished in compliance with the guidelines promulgated in the IEX
# API terms of service and manual
# See https://iextrading.com/api-exhibit-a/ for additional information
# and conditions of use


class IEX(_BaseReader):
    """
    Serves as the base class for all IEX API services.
    """

    _format = "json"

    def __init__(
        self, symbols=None, start=None, end=None, retry_count=3, pause=0.1, session=None
    ):
        super(IEX, self).__init__(
            symbols=symbols,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        )

    @property
    def service(self):
        """Service endpoint"""
        # This property will be overridden by the subclass
        raise NotImplementedError("IEX API service not specified.")

    @property
    def url(self):
        """API URL"""
        qstring = urlencode(self._get_params(self.symbols))
        return "https://api.iextrading.com/1.0/{}?{}".format(self.service, qstring)

    def read(self):
        """Read data"""
        df = super(IEX, self).read()
        if isinstance(df, pd.DataFrame):
            df = df.squeeze()
            if not isinstance(df, pd.DataFrame):
                df = pd.DataFrame(df)
        return df

    def _get_params(self, symbols):
        p = {}
        if isinstance(symbols, list):
            p["symbols"] = ",".join(symbols)
        elif isinstance(symbols, str):
            p["symbols"] = symbols
        return p

    def _output_error(self, out):
        """If IEX returns a non-200 status code, we need to notify the user of
        the error returned.

        Parameters
        ----------
        out: bytes
            The raw output from an HTTP request
        """
        try:
            content = json.loads(out.text)
        except Exception:
            raise TypeError("Failed to interpret response as JSON.")

        for key, string in content.items():
            e = "IEX Output error encountered: {}".format(string)
            if key == "error":
                raise Exception(e)

    def _read_lines(self, out):
        """IEX's output does not need anything complex, so we're overriding to
        use Pandas' default interpreter

        Parameters
        ----------
        out: bytes
            The raw output from an HTTP request

        Returns
        -------
        DataFrame
        """

        # IEX will return a blank line for invalid tickers:
        if isinstance(out, list):
            out = [x for x in out if x is not None]
        return pd.DataFrame(out) if len(out) > 0 else pd.DataFrame()



import datetime as dt

import pytest
import requests

from pandas_datareader import base as base

pytestmark = pytest.mark.stable


class TestBaseReader(object):
    def test_requests_not_monkey_patched(self):
        assert not hasattr(requests.Session(), "stor")

    def test_valid_retry_count(self):
        with pytest.raises(ValueError):
            base._BaseReader([], retry_count="stuff")
        with pytest.raises(ValueError):
            base._BaseReader([], retry_count=-1)

    def test_invalid_url(self):
        with pytest.raises(NotImplementedError):
            base._BaseReader([]).url

    def test_invalid_format(self):
        with pytest.raises(NotImplementedError):
            b = base._BaseReader([])
            b._format = "IM_NOT_AN_IMPLEMENTED_TYPE"
            b._read_one_data("a", None)

    def test_default_start_date(self):
        b = base._BaseReader([])
        assert b.default_start_date == dt.date.today() - dt.timedelta(days=365 * 5)


class TestDailyBaseReader(object):
    def test_get_params(self):
        with pytest.raises(NotImplementedError):
            b = base._DailyBaseReader()
            b._get_params()



import pandas as pd

from pandas_datareader.base import _BaseReader


class EcondbReader(_BaseReader):
    """Get data for the given name from Econdb."""

    _URL = "https://www.econdb.com/api/series/"
    _format = None
    _show = "labels"

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        params = dict(s.split("=") for s in self.symbols.split("&"))
        if "from" in params and not kwargs.get("start"):
            self.start = pd.to_datetime(params["from"], format="%Y-%m-%d")
        if "to" in params and not kwargs.get("end"):
            self.end = pd.to_datetime(params["to"], format="%Y-%m-%d")

    @property
    def url(self):
        """API URL"""
        if not isinstance(self.symbols, str):
            raise ValueError("data name must be string")

        return "{0}?{1}&format=json&page_size=500&expand=both".format(
            self._URL, self.symbols
        )

    def read(self):
        """read one data from specified URL"""
        results = self.session.get(self.url).json()["results"]
        df = pd.DataFrame({"dates": []}).set_index("dates")

        if self._show == "labels":

            def show_func(x):
                return x[x.find(":") + 1 :]

        elif self._show == "codes":

            def show_func(x):
                return x[: x.find(":")]

        for entry in results:
            series = pd.DataFrame(entry["data"])[["dates", "values"]].set_index("dates")
            head = entry["additional_metadata"]

            if head != "":  # this additional metadata is not blank
                series.columns = pd.MultiIndex.from_tuples(
                    [[show_func(x) for x in head.values()]],
                    names=[show_func(x) for x in head.keys()],
                )
            else:
                series.rename(columns={"values": entry["ticker"]}, inplace=True)

            if not df.empty:
                df = df.merge(series, how="outer", left_index=True, right_index=True)
            else:
                df = series
        if df.shape[0] > 0:
            df.index = pd.to_datetime(df.index, errors="ignore")
        df.index.name = "TIME_PERIOD"
        df = df.truncate(self.start, self.end)
        return df



from __future__ import division

import json
import re
import time

from pandas import DataFrame, isnull, notnull, to_datetime

from pandas_datareader._utils import RemoteDataError
from pandas_datareader.base import _DailyBaseReader
from pandas_datareader.yahoo.headers import DEFAULT_HEADERS


class YahooDailyReader(_DailyBaseReader):
    """
    Returns DataFrame of with historical over date range,
    start to end.
    To avoid being penalized by Yahoo! Finance servers, pauses between
    downloading 'chunks' of symbols can be specified.

    Parameters
    ----------
    symbols : string, array-like object (list, tuple, Series), or DataFrame
        Single stock symbol (ticker), array-like object of symbols or
        DataFrame with index containing stock symbols.
    start : string, int, date, datetime, Timestamp
        Starting date. Parses many different kind of date
        representations (e.g., 'JAN-01-2010', '1/1/10', 'Jan, 1, 1980'). Defaults to
        5 years before current date.
    end : string, int, date, datetime, Timestamp
        Ending date
    retry_count : int, default 3
        Number of times to retry query request.
    pause : int, default 0.1
        Time, in seconds, to pause between consecutive queries of chunks. If
        single value given for symbol, represents the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used. Passing a session
        is an advanced usage and you must set any required
        headers in the session directly.
    adjust_price : bool, default False
        If True, adjusts all prices in hist_data ('Open', 'High', 'Low',
        'Close') based on 'Adj Close' price. Adds 'Adj_Ratio' column and drops
        'Adj Close'.
    ret_index : bool, default False
        If True, includes a simple return index 'Ret_Index' in hist_data.
    chunksize : int, default 25
        Number of symbols to download consecutively before intiating pause.
    interval : string, default 'd'
        Time interval code, valid values are 'd' for daily, 'w' for weekly,
        'm' for monthly.
    get_actions : bool, default False
        If True, adds Dividend and Split columns to dataframe.
    adjust_dividends: bool, default true
        If True, adjusts dividends for splits.
    """

    def __init__(
        self,
        symbols=None,
        start=None,
        end=None,
        retry_count=3,
        pause=0.1,
        session=None,
        adjust_price=False,
        ret_index=False,
        chunksize=1,
        interval="d",
        get_actions=False,
        adjust_dividends=True,
    ):
        super().__init__(
            symbols=symbols,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            chunksize=chunksize,
        )

        # Ladder up the wait time between subsequent requests to improve
        # probability of a successful retry
        self.pause_multiplier = 2.5
        if session is None:
            self.headers = DEFAULT_HEADERS
        else:
            self.headers = session.headers

        self.adjust_price = adjust_price
        self.ret_index = ret_index
        self.interval = interval
        self._get_actions = get_actions

        if self.interval not in ["d", "wk", "mo", "m", "w"]:
            raise ValueError(
                "Invalid interval: valid values are  'd', 'wk' and 'mo'. 'm' and 'w' "
                "have been implemented for backward compatibility. 'v' has been moved "
                "to the yahoo-actions or yahoo-dividends APIs."
            )
        elif self.interval in ["m", "mo"]:
            self.pdinterval = "m"
            self.interval = "mo"
        elif self.interval in ["w", "wk"]:
            self.pdinterval = "w"
            self.interval = "wk"

        self.interval = "1" + self.interval
        self.adjust_dividends = adjust_dividends

    @property
    def get_actions(self):
        return self._get_actions

    @property
    def url(self):
        return "https://finance.yahoo.com/quote/{}/history"

    # Test test_get_data_interval() crashed because of this issue, probably
    # whole yahoo part of package wasn't
    # working properly
    def _get_params(self, symbol):
        # This needed because yahoo returns data shifted by 4 hours ago.
        four_hours_in_seconds = 14400
        unix_start = int(time.mktime(self.start.timetuple()))
        unix_start += four_hours_in_seconds
        day_end = self.end.replace(hour=23, minute=59, second=59)
        unix_end = int(time.mktime(day_end.timetuple()))
        unix_end += four_hours_in_seconds

        params = {
            "period1": unix_start,
            "period2": unix_end,
            "interval": self.interval,
            "frequency": self.interval,
            "filter": "history",
            "symbol": symbol,
        }
        return params

    def _read_one_data(self, url, params):
        """read one data from specified symbol"""

        symbol = params["symbol"]
        del params["symbol"]
        url = url.format(symbol)

        resp = self._get_response(url, params=params, headers=self.headers)
        ptrn = r"root\.App\.main = (.*?);\n}\(this\)\);"
        try:
            j = json.loads(re.search(ptrn, resp.text, re.DOTALL).group(1))
            data = j["context"]["dispatcher"]["stores"]["HistoricalPriceStore"]
        except KeyError:
            msg = "No data fetched for symbol {} using {}"
            raise RemoteDataError(msg.format(symbol, self.__class__.__name__))

        # price data
        prices = DataFrame(data["prices"])
        prices.columns = [col.capitalize() for col in prices.columns]
        prices["Date"] = to_datetime(to_datetime(prices["Date"], unit="s").dt.date)

        if "Data" in prices.columns:
            prices = prices[prices["Data"].isnull()]
        prices = prices[["Date", "High", "Low", "Open", "Close", "Volume", "Adjclose"]]
        prices = prices.rename(columns={"Adjclose": "Adj Close"})

        prices = prices.set_index("Date")
        prices = prices.sort_index().dropna(how="all")

        if self.ret_index:
            prices["Ret_Index"] = _calc_return_index(prices["Adj Close"])
        if self.adjust_price:
            prices = _adjust_prices(prices)

        # dividends & splits data
        if self.get_actions and data["eventsData"]:

            actions = DataFrame(data["eventsData"])
            actions.columns = [col.capitalize() for col in actions.columns]
            actions["Date"] = to_datetime(
                to_datetime(actions["Date"], unit="s").dt.date
            )

            types = actions["Type"].unique()
            if "DIVIDEND" in types:
                divs = actions[actions.Type == "DIVIDEND"].copy()
                divs = divs[["Date", "Amount"]].reset_index(drop=True)
                divs = divs.set_index("Date")
                divs = divs.rename(columns={"Amount": "Dividends"})
                prices = prices.join(divs, how="outer")

            if "SPLIT" in types:

                def split_ratio(row):
                    if float(row["Numerator"]) > 0:
                        if ":" in row["Splitratio"]:
                            n, m = row["Splitratio"].split(":")
                            return float(m) / float(n)
                        else:
                            return eval(row["Splitratio"])
                    else:
                        return 1

                splits = actions[actions.Type == "SPLIT"].copy()
                splits["SplitRatio"] = splits.apply(split_ratio, axis=1)
                splits = splits.reset_index(drop=True)
                splits = splits.set_index("Date")
                splits["Splits"] = splits["SplitRatio"]
                prices = prices.join(splits["Splits"], how="outer")

                if "DIVIDEND" in types and not self.adjust_dividends:
                    # dividends are adjusted automatically by Yahoo
                    adj = (
                        prices["Splits"].sort_index(ascending=False).fillna(1).cumprod()
                    )
                    prices["Dividends"] = prices["Dividends"] / adj

        return prices


def _adjust_prices(hist_data, price_list=None):
    """
    Return modifed DataFrame with adjusted prices based on
    'Adj Close' price. Adds 'Adj_Ratio' column.
    """
    if price_list is None:
        price_list = "Open", "High", "Low", "Close"
    adj_ratio = hist_data["Adj Close"] / hist_data["Close"]

    data = hist_data.copy()
    for item in price_list:
        data[item] = hist_data[item] * adj_ratio
    data["Adj_Ratio"] = adj_ratio
    del data["Adj Close"]
    return data


def _calc_return_index(price_df):
    """
    Return a returns index from a input price df or series. Initial value
    (typically NaN) is set to 1.
    """
    df = price_df.pct_change().add(1).cumprod()
    mask = notnull(df.iloc[1]) & isnull(df.iloc[0])
    if mask:
        df.loc[df.index[0]] = 1

    # Check for first stock listings after starting date of index in ret_index
    # If True, find first_valid_index and set previous entry to 1.
    if not mask:
        tstamp = df.first_valid_index()
        t_idx = df.index.get_loc(tstamp) - 1
        df.iloc[t_idx] = 1

    return df



from pandas_datareader import data as web
from pandas_datareader._testing import skip_on_exception
from pandas_datareader._utils import RemoteDataError


class TestNasdaqSymbols(object):
    @skip_on_exception(RemoteDataError)
    def test_get_symbols(self):
        symbols = web.DataReader("symbols", "nasdaq")
        assert "IBM" in symbols.index



# pylint: disable-msg=E1101,W0613,W0603

from __future__ import unicode_literals

from collections import OrderedDict
import itertools
import re
import sys

import numpy as np
import pandas as pd

from pandas_datareader.io.util import _read_content


def read_jsdmx(path_or_buf):
    """
    Convert a SDMX-JSON string to pandas object

    Parameters
    ----------
    path_or_buf : a valid SDMX-JSON string or file-like
        https://github.com/sdmx-twg/sdmx-json

    Returns
    -------
    results : Series, DataFrame, or dictionary of Series or DataFrame.
    """

    jdata = _read_content(path_or_buf)

    try:
        import simplejson as json
    except ImportError:
        if sys.version_info[:2] < (2, 7):
            raise ImportError("simplejson is required in python 2.6")
        import json

    if isinstance(jdata, dict):
        data = jdata
    else:
        data = json.loads(jdata, object_pairs_hook=OrderedDict)

    structure = data["structure"]
    index = _parse_dimensions(structure["dimensions"]["observation"])
    columns = _parse_dimensions(structure["dimensions"]["series"])

    dataset = data["dataSets"]
    if len(dataset) != 1:
        raise ValueError("length of 'dataSets' must be 1")
    dataset = dataset[0]
    values = _parse_values(dataset, index=index, columns=columns)

    df = pd.DataFrame(values, columns=columns, index=index)
    return df


def _get_indexer(index):
    if index.nlevels == 1:
        return [str(i) for i in range(len(index))]
    else:
        it = itertools.product(*[range(len(level)) for level in index.levels])
        return [":".join(map(str, i)) for i in it]


def _fix_quarter_values(value):
    """Make raw quarter values Pandas-friendly (e.g. 'Q4-2018' -> '2018Q4')."""
    m = re.match(r"Q([1-4])-(\d\d\d\d)", value)
    if not m:
        return value
    quarter, year = m.groups()
    value = "%sQ%s" % (quarter, year)
    return value


def _parse_values(dataset, index, columns):
    size = len(index)
    series = dataset["series"]

    values = []
    # for s_key, s_value in iteritems(series):
    for s_key in _get_indexer(columns):
        try:
            observations = series[s_key]["observations"]
            observed = []
            for o_key in _get_indexer(index):
                try:
                    observed.append(observations[o_key][0])
                except KeyError:
                    observed.append(np.nan)
        except KeyError:
            observed = [np.nan] * size

        values.append(observed)

    return np.transpose(np.array(values))


def _parse_dimensions(dimensions):
    arrays = []
    names = []
    for key in dimensions:
        values = [v["name"] for v in key["values"]]

        role = key.get("role", None)
        if role in ("time", "TIME_PERIOD"):
            values = [_fix_quarter_values(v) for v in values]
            values = pd.DatetimeIndex(values)

        arrays.append(values)
        names.append(key["name"])
    midx = pd.MultiIndex.from_product(arrays, names=names)
    if len(arrays) == 1 and isinstance(midx, pd.MultiIndex):
        # Fix for pandas >= 0.21
        midx = midx.levels[0]

    return midx



"""
Default header
"""
DEFAULT_HEADERS = {
    "Connection": "keep-alive",
    "Expires": str(-1),
    "Upgrade-Insecure-Requests": str(1),
    # Google Chrome:
    "User-Agent": (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
        "(KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"
    ),
}



import os

import pandas as pd
import pytest

from pandas_datareader import data as web
from pandas_datareader.compat import assert_frame_equal

TEST_API_KEY = os.getenv("QUANDL_API_KEY")
# Ensure blank TEST_API_KEY not used in pull request
TEST_API_KEY = None if not TEST_API_KEY else TEST_API_KEY

pytestmark = [
    pytest.mark.requires_api_key,
    pytest.mark.quandl,
    pytest.mark.skipif(TEST_API_KEY is None, reason="QUANDL_API_KEY not set"),
]


class TestQuandl(object):
    # we test data from 10 years back where possible, 2 years otherwise, or...
    start10 = "2007-01-01"  # over ten years back
    end10 = "2007-01-05"
    day10 = "2007-01-04"
    start2 = "2015-01-01"  # over two years back
    end2 = "2015-01-05"
    day2 = "2015-01-02"

    def check_headers(self, df, expected_cols):
        expected_cols = frozenset(expected_cols)
        act_cols = frozenset(df.columns.tolist())
        assert expected_cols == act_cols, "unexpected cols: " + str(act_cols)

    def test_db_wiki_us(self):
        df = web.DataReader(
            "F", "quandl", self.start10, self.end10, api_key=TEST_API_KEY
        )
        self.check_headers(
            df,
            [
                "Open",
                "High",
                "Low",
                "Close",
                "Volume",
                "ExDividend",
                "SplitRatio",
                "AdjOpen",
                "AdjHigh",
                "AdjLow",
                "AdjClose",
                "AdjVolume",
            ],
        )
        assert df.Close.at[pd.to_datetime(self.day10)] == 7.70

    def test_db_fse_frankfurt(self):
        # ALV_X: Allianz SE
        df = web.DataReader(
            "FSE/ALV_X", "quandl", self.start10, self.end10, api_key=TEST_API_KEY
        )
        self.check_headers(
            df,
            [
                "Open",
                "High",
                "Low",
                "Close",
                "Change",
                "TradedVolume",
                "Turnover",
                "LastPriceoftheDay",
                "DailyTradedUnits",
                "DailyTurnover",
            ],
        )
        assert df.Close.at[pd.to_datetime(self.day10)] == 159.45

    def test_fse_eon(self):
        # EON_X: E.on Se
        df = web.DataReader(
            "FSE/EON_X", "quandl", self.start2, self.end2, api_key=TEST_API_KEY
        )
        self.check_headers(
            df,
            [
                "Low",
                "LastPriceoftheDay",
                "Turnover",
                "Open",
                "DailyTurnover",
                "TradedVolume",
                "Change",
                "DailyTradedUnits",
                "Close",
                "High",
            ],
        )
        assert df.Close.at[pd.to_datetime(self.day2)] == 14.03

    def test_db_euronext_be_fr_nl_pt(self):
        # FP: Total SA
        # as of 2017-06-11, some datasets end a few months after their start,
        # e.g. ALVD, BASD
        df = web.DataReader(
            "EURONEXT/FP", "quandl", self.start2, self.end2, api_key=TEST_API_KEY
        )
        self.check_headers(df, ["Open", "High", "Low", "Last", "Turnover", "Volume"])
        assert df.Last.at[pd.to_datetime(self.day2)] == 42.525
        df2 = web.DataReader("FP.FR", "quandl", self.start2, self.end2)
        assert (df.Last == df2.Last).all()

    def test_hk_hsbc_uk(self):
        # 00005: HSBC
        df = web.DataReader(
            "HKEX/00005", "quandl", self.start2, self.end2, api_key=TEST_API_KEY
        )
        self.check_headers(
            df,
            [
                "NominalPrice",
                "NetChange",
                "Change",
                "Bid",
                "Ask",
                "PEx",
                "High",
                "Low",
                "PreviousClose",
                "ShareVolume000",
                "Turnover000",
                "LotSize",
            ],
        )
        # as of 2017-06-11, Price == LastClose, all others are NaN
        assert df.NominalPrice.at[pd.to_datetime(self.day2)] == 74.0

    def test_db_nse_in(self):
        # TCS: Tata Consutancy Services
        df = web.DataReader(
            "NSE/TCS", "quandl", self.start10, self.end10, api_key=TEST_API_KEY
        )
        self.check_headers(
            df,
            [
                "Open",
                "High",
                "Low",
                "Last",
                "Close",
                "TotalTradeQuantity",
                "TurnoverLacs",
            ],
        )
        assert df.Close.at[pd.to_datetime(self.day10)] == 1259.05

    def test_db_tse_jp(self):
        # TSE/6758: Sony Corp.
        df = web.DataReader(
            "TSE/6758", "quandl", self.start10, self.end10, api_key=TEST_API_KEY
        )
        self.check_headers(df, ["Open", "High", "Low", "Close", "Volume"])
        assert df.Close.at[pd.to_datetime(self.day10)] == 5190.0

        df2 = web.get_data_quandl(
            "TSE/6758", self.start10, self.end10, api_key=TEST_API_KEY
        )
        assert_frame_equal(df, df2)

    def test_db_hkex_cn(self):
        # HKEX/00941: China Mobile
        df = web.DataReader(
            "HKEX/00941", "quandl", self.start2, self.end2, api_key=TEST_API_KEY
        )
        self.check_headers(
            df,
            [
                "NominalPrice",
                "NetChange",
                "Change",
                "Bid",
                "Ask",
                "PEx",
                "High",
                "Low",
                "PreviousClose",
                "ShareVolume000",
                "Turnover000",
                "LotSize",
            ],
        )
        assert df.High.at[pd.to_datetime(self.day2)] == 91.9



from datetime import datetime

import numpy as np
import pandas as pd
from pandas import DataFrame, testing as tm
import pytest
import requests
from requests.exceptions import ConnectionError

from pandas_datareader import data as web
from pandas_datareader._testing import skip_on_exception
from pandas_datareader._utils import RemoteDataError
from pandas_datareader.data import YahooDailyReader

XFAIL_REASON = "Known connection failures on Yahoo when testing!"

pytestmark = pytest.mark.stable


class TestYahoo(object):
    @classmethod
    def setup_class(cls):
        pytest.importorskip("lxml")

    @skip_on_exception(RemoteDataError)
    def test_yahoo(self):
        # Asserts that yahoo is minimally working
        start = datetime(2010, 1, 1)
        end = datetime(2013, 1, 25)

        assert round(web.DataReader("F", "yahoo", start, end)["Close"][-1], 2) == 13.68

    def test_yahoo_fails(self):
        start = datetime(2010, 1, 1)
        end = datetime(2013, 1, 27)

        with pytest.raises(Exception):
            web.DataReader("NON EXISTENT TICKER", "yahoo", start, end)

    def test_get_quote_series(self):
        stringlist = ["GOOG", "AAPL"]
        fields = ["exchange", "sharesOutstanding", "epsForward"]
        try:
            AAPL = web.get_quote_yahoo("AAPL")
            df = web.get_quote_yahoo(pd.Series(stringlist))
        except ConnectionError:
            pytest.xfail(reason=XFAIL_REASON)
        tm.assert_series_equal(AAPL.iloc[0][fields], df.loc["AAPL"][fields])
        assert sorted(stringlist) == sorted(list(df.index.values))

    def test_get_quote_string(self):
        try:
            df = web.get_quote_yahoo("GOOG")
        except ConnectionError:
            pytest.xfail(reason=XFAIL_REASON)

        assert not pd.isnull(df["marketCap"][0])

    def test_get_quote_stringlist(self):
        stringlist = ["GOOG", "AAPL"]
        try:
            df = web.get_quote_yahoo(stringlist)
        except ConnectionError:
            pytest.xfail(reason=XFAIL_REASON)
        assert sorted(stringlist) == sorted(list(df.index.values))

    def test_get_quote_comma_name(self):
        try:
            df = web.get_quote_yahoo(["RGLD"])
        except ConnectionError:
            pytest.xfail(reason=XFAIL_REASON)
        assert df["longName"][0] == "Royal Gold, Inc."

    @pytest.mark.skip(
        "Unreliable test, receive partial " "components back for dow_jones"
    )
    def test_get_components_dow_jones(self):  # pragma: no cover
        df = web.get_components_yahoo("^DJI")  # Dow Jones
        assert isinstance(df, pd.DataFrame)
        assert len(df) == 30

    @pytest.mark.skip("Unreliable test, receive partial " "components back for dax")
    def test_get_components_dax(self):  # pragma: no cover
        df = web.get_components_yahoo("^GDAXI")  # DAX
        assert isinstance(df, pd.DataFrame)

        assert len(df) == 30
        assert df[df.name.str.contains("adidas", case=False)].index == "ADS.DE"

    @pytest.mark.skip(
        "Unreliable test, receive partial " "components back for nasdaq_100"
    )
    def test_get_components_nasdaq_100(self):  # pragma: no cover
        # As of 7/12/13, the conditional will
        # return false because the link is invalid

        df = web.get_components_yahoo("^NDX")  # NASDAQ-100
        assert isinstance(df, pd.DataFrame)

        if len(df) > 1:
            # Usual culprits, should be around for a while
            assert "AAPL" in df.index
            assert "GOOG" in df.index
            assert "AMZN" in df.index
        else:
            expected = DataFrame({"exchange": "N/A", "name": "@^NDX"}, index=["@^NDX"])
            tm.assert_frame_equal(df, expected)

    @skip_on_exception(RemoteDataError)
    def test_get_data_single_symbol(self):
        # single symbol
        # http://finance.yahoo.com/q/hp?s=GOOG&a=09&b=08&c=2010&d=09&e=10&f=2010&g=d
        # just test that we succeed
        web.get_data_yahoo("GOOG")

    @skip_on_exception(RemoteDataError)
    def test_data_with_no_actions(self):
        web.get_data_yahoo("TSLA")

    @skip_on_exception(RemoteDataError)
    def test_get_data_adjust_price(self):
        goog = web.get_data_yahoo("GOOG")
        goog_adj = web.get_data_yahoo("GOOG", adjust_price=True)
        assert "Adj Close" not in goog_adj.columns
        assert (goog["Open"] * goog_adj["Adj_Ratio"]).equals(goog_adj["Open"])

    @pytest.mark.xfail(reason="Yahoo are returning an extra day 31st Dec 2012")
    def test_get_data_interval(self):
        # daily interval data
        pan = web.get_data_yahoo("XOM", "2013-01-01", "2013-12-31", interval="d")
        assert len(pan) == 252

        # weekly interval data
        pan = web.get_data_yahoo("XOM", "2013-01-01", "2013-12-31", interval="w")
        assert len(pan) == 53

        # monthly interval data
        pan = web.get_data_yahoo("XOM", "2012-12-31", "2013-12-31", interval="m")
        assert len(pan) == 12

        # test fail on invalid interval
        with pytest.raises(ValueError):
            web.get_data_yahoo("XOM", interval="NOT VALID")

    @skip_on_exception(RemoteDataError)
    def test_get_data_multiple_symbols(self):
        # just test that we succeed
        sl = ["AAPL", "AMZN", "GOOG"]
        web.get_data_yahoo(sl, "2012")

    @pytest.mark.parametrize("adj_pr", [True, False])
    @skip_on_exception(RemoteDataError)
    def test_get_data_null_as_missing_data(self, adj_pr):
        result = web.get_data_yahoo("SRCE", "20160626", "20160705", adjust_price=adj_pr)
        # sanity checking
        floats = ["Open", "High", "Low", "Close"]
        if adj_pr:
            floats.append("Adj_Ratio")
        else:
            floats.append("Adj Close")

        assert result[floats].dtypes.all() == np.float64

    @skip_on_exception(RemoteDataError)
    def test_get_data_multiple_symbols_two_dates(self):
        data = web.get_data_yahoo(["GE", "MSFT", "INTC"], "JAN-01-12", "JAN-31-12")
        result = data.Close.loc["01-18-12"].T
        assert result.size == 3

        # sanity checking
        assert result.dtypes == np.float64

        expected = np.array(
            [
                [18.99, 28.4, 25.18],
                [18.58, 28.31, 25.13],
                [19.03, 28.16, 25.52],
                [18.81, 28.82, 25.87],
            ]
        )
        df = data.Open
        result = df[(df.index >= "Jan-15-12") & (df.index <= "Jan-20-12")]
        assert expected.shape == result.shape

    def test_get_date_ret_index(self):
        pan = web.get_data_yahoo(["GE", "INTC", "IBM"], "1977", "1987", ret_index=True)
        assert hasattr(pan, "Ret_Index")

        if hasattr(pan, "Ret_Index") and hasattr(pan.Ret_Index, "INTC"):
            tstamp = pan.Ret_Index.INTC.first_valid_index()
            result = pan.Ret_Index.loc[tstamp, "INTC"]
            assert result == 1.0

        # sanity checking
        assert np.issubdtype(pan.values.dtype, np.floating)

    def test_get_data_yahoo_actions(self):
        start = datetime(1990, 1, 1)
        end = datetime(2018, 4, 5)

        actions = web.get_data_yahoo_actions("AAPL", start, end, adjust_dividends=False)

        assert sum(actions["action"] == "DIVIDEND") == 47
        assert sum(actions["action"] == "SPLIT") == 3

        assert actions.loc["2005-02-28", "action"][0] == "SPLIT"
        assert actions.loc["2005-02-28", "value"][0] == 1 / 2.0

        assert actions.loc["1995-11-21", "action"][0] == "DIVIDEND"
        assert round(actions.loc["1995-11-21", "value"][0], 3) == 0.030

        actions = web.get_data_yahoo_actions("AAPL", start, end, adjust_dividends=True)

        assert actions.loc["1995-11-21", "action"][0] == "DIVIDEND"
        assert round(actions.loc["1995-11-21", "value"][0], 4) == 0.0011

    def test_get_data_yahoo_actions_invalid_symbol(self):
        start = datetime(1990, 1, 1)
        end = datetime(2000, 4, 5)

        with pytest.raises(IOError):
            web.get_data_yahoo_actions("UNKNOWN TICKER", start, end)

    @skip_on_exception(RemoteDataError)
    def test_yahoo_reader_class(self):
        r = YahooDailyReader("GOOG", start="JAN-01-2015")
        df = r.read()

        assert df.Volume.loc["JAN-02-2015"] == 1447563

        session = requests.Session()

        r = YahooDailyReader("GOOG", session=session)
        assert r.session is session

    def test_yahoo_datareader(self):
        start = datetime(2010, 1, 1)
        end = datetime(2015, 5, 9)
        # yahoo will adjust for dividends by default
        result = web.DataReader("AAPL", "yahoo-actions", start, end)

        exp_idx = pd.DatetimeIndex(
            [
                "2015-05-07",
                "2015-02-05",
                "2014-11-06",
                "2014-08-07",
                "2014-06-09",
                "2014-05-08",
                "2014-02-06",
                "2013-11-06",
                "2013-08-08",
                "2013-05-09",
                "2013-02-07",
                "2012-11-07",
                "2012-08-09",
            ]
        )

        exp = pd.DataFrame(
            {
                "action": [
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                    "SPLIT",
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                ],
                "value": [
                    0.130000,
                    0.117500,
                    0.117500,
                    0.117500,
                    0.142857,
                    0.117500,
                    0.108929,
                    0.108929,
                    0.108929,
                    0.108929,
                    0.094643,
                    0.094643,
                    0.094643,
                ],
            },
            index=exp_idx,
        )
        exp.index.name = "Date"
        tm.assert_frame_equal(result.reindex_like(exp).round(2), exp.round(2))

        # where dividends are not adjusted for splits
        result = web.get_data_yahoo_actions("AAPL", start, end, adjust_dividends=False)

        exp = pd.DataFrame(
            {
                "action": [
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                    "SPLIT",
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                    "DIVIDEND",
                ],
                "value": [
                    0.1300,
                    0.1175,
                    0.1175,
                    0.1175,
                    0.1429,
                    0.8225,
                    0.7625,
                    0.7625,
                    0.7625,
                    0.7625,
                    0.6625,
                    0.6625,
                    0.6625,
                ],
            },
            index=exp_idx,
        )
        exp.index.name = "Date"
        tm.assert_frame_equal(result.reindex_like(exp).round(4), exp.round(4))

        # test cases with "1/0" split ratio in actions -
        # no split, just chnage symbol from POT to NTR
        start = datetime(2017, 12, 30)
        end = datetime(2018, 12, 30)

        result = web.DataReader("NTR", "yahoo-actions", start, end)

        exp_idx = pd.DatetimeIndex(
            ["2018-12-28", "2018-09-27", "2018-06-28", "2018-03-28"]
        )

        exp = pd.DataFrame(
            {
                "action": ["DIVIDEND", "DIVIDEND", "DIVIDEND", "DIVIDEND"],
                "value": [0.43, 0.40, 0.40, 0.40],
            },
            index=exp_idx,
        )
        exp.index.name = "Date"
        tm.assert_frame_equal(result.reindex_like(exp).round(2), exp.round(2))

    @skip_on_exception(RemoteDataError)
    def test_yahoo_DataReader_multi(self):
        start = datetime(2010, 1, 1)
        end = datetime(2015, 5, 9)
        result = web.DataReader(["AAPL", "F"], "yahoo-actions", start, end)
        assert isinstance(result, dict)



from datetime import datetime
import os

import numpy as np
import pandas as pd
from pandas import testing as tm
import pytest

from pandas_datareader import data as web


@pytest.fixture
def aapl():
    aapl = web.Options("aapl", "yahoo")
    yield aapl
    aapl.close()


@pytest.fixture
def month():

    # AAPL has monthlies
    today = datetime.today()
    month = today.month + 1

    if month > 12:  # pragma: no cover
        month = 1

    return month


@pytest.fixture
def year():

    # AAPL has monthlies
    today = datetime.today()
    year = today.year
    month = today.month + 1

    if month > 12:  # pragma: no cover
        year = year + 1

    return year


@pytest.fixture
def expiry(month, year):
    return datetime(year, month, 1)


@pytest.fixture
def json1(datapath):
    dirpath = datapath("yahoo", "data")
    json1 = "file://" + os.path.join(dirpath, "yahoo_options1.json")
    return json1


@pytest.fixture
def json2(datapath):
    # see gh-22: empty table
    dirpath = datapath("yahoo", "data")
    json2 = "file://" + os.path.join(dirpath, "yahoo_options2.json")
    return json2


@pytest.fixture
def data1(aapl, json1):
    return aapl._process_data(aapl._parse_url(json1))


class TestYahooOptions(object):
    @classmethod
    def setup_class(cls):
        pytest.skip("Skip all Yahoo! tests.")

    def assert_option_result(self, df):
        """
        Validate returned option data has expected format.
        """
        assert isinstance(df, pd.DataFrame)
        assert len(df) > 1

        exp_columns = pd.Index(
            [
                "Last",
                "Bid",
                "Ask",
                "Chg",
                "PctChg",
                "Vol",
                "Open_Int",
                "IV",
                "Root",
                "IsNonstandard",
                "Underlying",
                "Underlying_Price",
                "Quote_Time",
                "Last_Trade_Date",
                "JSON",
            ]
        )
        tm.assert_index_equal(df.columns, exp_columns)
        assert df.index.names == [u"Strike", u"Expiry", u"Type", u"Symbol"]

        dtypes = [
            np.dtype(x)
            for x in ["float64"] * 7
            + [
                "float64",
                "object",
                "bool",
                "object",
                "float64",
                "datetime64[ns]",
                "datetime64[ns]",
                "object",
            ]
        ]
        tm.assert_series_equal(df.dtypes, pd.Series(dtypes, index=exp_columns))

    def test_get_options_data(self, aapl, expiry):
        # see gh-6105: regression test
        with pytest.raises(ValueError):
            aapl.get_options_data(month=3)

        with pytest.raises(ValueError):
            aapl.get_options_data(year=1992)

        options = aapl.get_options_data(expiry=expiry)
        self.assert_option_result(options)

    def test_get_near_stock_price(self, aapl, expiry):
        options = aapl.get_near_stock_price(call=True, put=True, expiry=expiry)
        self.assert_option_result(options)

    def test_options_is_not_none(self):
        option = web.Options("aapl", "yahoo")
        assert option is not None

    def test_get_call_data(self, aapl, expiry):
        calls = aapl.get_call_data(expiry=expiry)

        self.assert_option_result(calls)
        assert calls.index.levels[2][0] == "call"

    def test_get_put_data(self, aapl, expiry):
        puts = aapl.get_put_data(expiry=expiry)

        self.assert_option_result(puts)
        assert puts.index.levels[2][1] == "put"

    def test_get_expiry_dates(self, aapl):
        dates = aapl._get_expiry_dates()
        assert len(dates) > 1

    def test_get_all_data(self, aapl):
        data = aapl.get_all_data(put=True)

        assert len(data) > 1
        self.assert_option_result(data)

    def test_get_data_with_list(self, aapl):
        data = aapl.get_call_data(expiry=aapl.expiry_dates)

        assert len(data) > 1
        self.assert_option_result(data)

    def test_get_all_data_calls_only(self, aapl):
        data = aapl.get_all_data(call=True, put=False)

        assert len(data) > 1
        self.assert_option_result(data)

    def test_get_underlying_price(self, aapl):
        # see gh-7
        options_object = web.Options("^spxpm", "yahoo")
        quote_price = options_object.underlying_price

        assert isinstance(quote_price, float)

        # Tests the weekend quote time format
        price, quote_time = aapl.underlying_price, aapl.quote_time

        assert isinstance(price, (int, float, complex))
        assert isinstance(quote_time, (datetime, pd.Timestamp))

    @pytest.mark.xfail(reason="Invalid URL scheme")
    def test_chop(self, aapl, data1):
        # gh-7625: regression test
        aapl._chop_data(data1, above_below=2, underlying_price=np.nan)
        chopped = aapl._chop_data(data1, above_below=2, underlying_price=100)

        assert isinstance(chopped, pd.DataFrame)
        assert len(chopped) > 1

        chopped2 = aapl._chop_data(data1, above_below=2, underlying_price=None)

        assert isinstance(chopped2, pd.DataFrame)
        assert len(chopped2) > 1

    @pytest.mark.xfail(reason="Invalid URL scheme")
    def test_chop_out_of_strike_range(self, aapl, data1):
        # gh-7625: regression test
        aapl._chop_data(data1, above_below=2, underlying_price=np.nan)
        chopped = aapl._chop_data(data1, above_below=2, underlying_price=100000)

        assert isinstance(chopped, pd.DataFrame)
        assert len(chopped) > 1

    @pytest.mark.xfail(reason="Invalid URL scheme")
    def test_sample_page_chg_float(self, data1):
        # Tests that numeric columns with comma's are appropriately dealt with
        assert data1["Chg"].dtype == "float64"

    def test_month_year(self, aapl, month, year):
        # see gh-168
        data = aapl.get_call_data(month=month, year=year)

        assert len(data) > 1
        assert data.index.levels[0].dtype == "float64"

        self.assert_option_result(data)

    @pytest.mark.xfail(reason="Invalid URL scheme")
    def test_empty_table(self, aapl, json2):
        # see gh-22
        empty = aapl._process_data(aapl._parse_url(json2))
        assert len(empty) == 0



from pandas_datareader.base import _DailyBaseReader


class StooqDailyReader(_DailyBaseReader):

    """
    Returns DataFrame/dict of Dataframes of historical stock prices from
    symbols, over date range, start to end.

    Parameters
    ----------
    symbols : string, array-like object (list, tuple, Series), or DataFrame
        Single stock symbol (ticker), array-like object of symbols or
        DataFrame with index containing stock symbols.
    start : string, int, date, datetime, Timestamp
        Starting date. Parses many different kind of date
        representations (e.g., 'JAN-01-2010', '1/1/10', 'Jan, 1, 1980'). Defaults to
        20 years before current date.
    end : string, int, date, datetime, Timestamp
        Ending date
    retry_count : int, default 3
        Number of times to retry query request.
    pause : int, default 0.1
        Time, in seconds, to pause between consecutive queries of chunks. If
        single value given for symbol, represents the pause between retries.
    chunksize : int, default 25
        Number of symbols to download consecutively before initiating pause.
    session : Session, default None
        requests.sessions.Session instance to be used
    freq: string, d, w, m ,q, y for daily, weekly, monthly, quarterly, yearly

    Notes
    -----
    See `Stooq <https://stooq.com>`__
    """

    @property
    def url(self):
        """API URL"""
        return "https://stooq.com/q/d/l/"

    def _get_params(self, symbol, country="US"):
        symbol_parts = symbol.split(".")
        if not symbol.startswith("^"):
            if len(symbol_parts) == 1:
                symbol = ".".join([symbol, country])
            elif symbol_parts[1].lower() == "pl":
                symbol = symbol_parts[0]
            else:
                if symbol_parts[1].lower() not in [
                    "de",
                    "hk",
                    "hu",
                    "jp",
                    "uk",
                    "us",
                    "f",
                    "b",
                ]:
                    symbol = ".".join([symbol, "US"])

        params = {
            "s": symbol,
            "i": self.freq or "d",
            "d1": self.start.strftime("%Y%m%d"),
            "d2": self.end.strftime("%Y%m%d"),
        }

        return params



import pandas as pd

from pandas_datareader._utils import RemoteDataError
from pandas_datareader.av import AlphaVantage


class AVForexReader(AlphaVantage):
    """
    Returns DataFrame of the Alpha Vantage Foreign Exchange (FX) Exchange Rates
    data.

    .. versionadded:: 0.7.0

    Parameters
    ----------
    symbols : string, array-like object (list, tuple, Series)
        Single currency pair (formatted 'FROM/TO') or list of the same.
    retry_count : int, default 3
        Number of times to retry query request.
    pause : int, default 0.1
        Time, in seconds, to pause between consecutive queries of chunks. If
        single value given for symbol, represents the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used
    api_key : str, optional
        Alpha Vantage API key . If not provided the environmental variable
        ALPHAVANTAGE_API_KEY is read. The API key is *required*.
    """

    def __init__(
        self, symbols=None, retry_count=3, pause=0.1, session=None, api_key=None
    ):

        super(AVForexReader, self).__init__(
            symbols=symbols,
            start=None,
            end=None,
            retry_count=retry_count,
            pause=pause,
            session=session,
            api_key=api_key,
        )
        self.from_curr = {}
        self.to_curr = {}
        self.optional_params = {}
        if isinstance(symbols, str):
            self.symbols = [symbols]
        else:
            self.symbols = symbols
        try:
            for pair in self.symbols:
                self.from_curr[pair] = pair.split("/")[0]
                self.to_curr[pair] = pair.split("/")[1]
        except Exception as e:
            print(e)
            raise ValueError(
                "Please input a currency pair "
                "formatted 'FROM/TO' or a list of "
                "currency symbols"
            )

    @property
    def function(self):
        return "CURRENCY_EXCHANGE_RATE"

    @property
    def data_key(self):
        return "Realtime Currency Exchange Rate"

    @property
    def params(self):
        params = {"apikey": self.api_key, "function": self.function}
        params.update(self.optional_params)
        return params

    def read(self):
        result = []
        for pair in self.symbols:
            self.optional_params = {
                "from_currency": self.from_curr[pair],
                "to_currency": self.to_curr[pair],
            }
            data = super(AVForexReader, self).read()
            result.append(data)
        df = pd.concat(result, axis=1)
        df.columns = self.symbols
        return df

    def _read_lines(self, out):
        try:
            df = pd.DataFrame.from_dict(out[self.data_key], orient="index")
        except KeyError:
            raise RemoteDataError()
        df.sort_index(ascending=True, inplace=True)
        df.index = [id[3:] for id in df.index]
        return df



"""
Module contains tools for collecting data from various remote sources
"""

# flake8: noqa

import warnings

from pandas.util._decorators import deprecate_kwarg

from pandas_datareader.av.forex import AVForexReader
from pandas_datareader.av.quotes import AVQuotesReader
from pandas_datareader.av.sector import AVSectorPerformanceReader
from pandas_datareader.av.time_series import AVTimeSeriesReader
from pandas_datareader.bankofcanada import BankOfCanadaReader
from pandas_datareader.econdb import EcondbReader
from pandas_datareader.enigma import EnigmaReader
from pandas_datareader.eurostat import EurostatReader
from pandas_datareader.exceptions import DEP_ERROR_MSG, ImmediateDeprecationError
from pandas_datareader.famafrench import FamaFrenchReader
from pandas_datareader.fred import FredReader
from pandas_datareader.iex.daily import IEXDailyReader
from pandas_datareader.iex.deep import Deep as IEXDeep
from pandas_datareader.iex.tops import LastReader as IEXLasts, TopsReader as IEXTops
from pandas_datareader.moex import MoexReader
from pandas_datareader.nasdaq_trader import get_nasdaq_symbols
from pandas_datareader.naver import NaverDailyReader
from pandas_datareader.oecd import OECDReader
from pandas_datareader.quandl import QuandlReader
from pandas_datareader.stooq import StooqDailyReader
from pandas_datareader.tiingo import (
    TiingoDailyReader,
    TiingoIEXHistoricalReader,
    TiingoQuoteReader,
)
from pandas_datareader.yahoo.actions import YahooActionReader, YahooDivReader
from pandas_datareader.yahoo.components import _get_data as get_components_yahoo
from pandas_datareader.yahoo.daily import YahooDailyReader
from pandas_datareader.yahoo.options import Options as YahooOptions
from pandas_datareader.yahoo.quotes import YahooQuotesReader

__all__ = [
    "get_components_yahoo",
    "get_data_enigma",
    "get_data_famafrench",
    "get_data_fred",
    "get_data_moex",
    "get_data_quandl",
    "get_data_yahoo",
    "get_data_yahoo_actions",
    "get_nasdaq_symbols",
    "get_quote_yahoo",
    "get_tops_iex",
    "get_summary_iex",
    "get_records_iex",
    "get_recent_iex",
    "get_markets_iex",
    "get_last_iex",
    "get_iex_symbols",
    "get_iex_book",
    "get_dailysummary_iex",
    "get_data_stooq",
    "DataReader",
]


def get_data_alphavantage(*args, **kwargs):
    return AVTimeSeriesReader(*args, **kwargs).read()


def get_data_fred(*args, **kwargs):
    return FredReader(*args, **kwargs).read()


def get_data_famafrench(*args, **kwargs):
    return FamaFrenchReader(*args, **kwargs).read()


def get_data_yahoo(*args, **kwargs):
    return YahooDailyReader(*args, **kwargs).read()


def get_data_enigma(*args, **kwargs):
    return EnigmaReader(*args, **kwargs).read()


def get_quote_av(*args, **kwargs):
    return AVQuotesReader(*args, **kwargs).read()


def get_data_yahoo_actions(*args, **kwargs):
    return YahooActionReader(*args, **kwargs).read()


def get_quote_yahoo(*args, **kwargs):
    return YahooQuotesReader(*args, **kwargs).read()


def get_data_quandl(*args, **kwargs):
    return QuandlReader(*args, **kwargs).read()


def get_data_moex(*args, **kwargs):
    return MoexReader(*args, **kwargs).read()


def get_data_stooq(*args, **kwargs):
    return StooqDailyReader(*args, **kwargs).read()


def get_tops_iex(*args, **kwargs):
    return IEXTops(*args, **kwargs).read()


def get_last_iex(*args, **kwargs):
    return IEXLasts(*args, **kwargs).read()


def get_data_tiingo(*args, **kwargs):
    return TiingoDailyReader(*args, **kwargs).read()


def get_iex_data_tiingo(*args, **kwargs):
    return TiingoIEXHistoricalReader(*args, **kwargs).read()


def get_quotes_tiingo(*args, **kwargs):
    return TiingoQuoteReader(*args, **kwargs).read()


def get_exchange_rate_av(*args, **kwargs):
    return AVForexReader(*args, **kwargs).read()


def get_sector_performance_av(*args, **kwargs):
    return AVSectorPerformanceReader(*args, **kwargs).read()


def get_markets_iex(*args, **kwargs):
    """
    Returns near-real time volume data across markets segregated by tape
    and including a percentage of overall volume during the session

    This endpoint does not accept any parameters.

    Reference: https://www.iextrading.com/developer/docs/#markets

    Returns
    -------
    DataFrame
    """
    from pandas_datareader.iex.market import MarketReader

    return MarketReader(*args, **kwargs).read()


def get_dailysummary_iex(*args, **kwargs):
    """
    Returns a summary of daily market volume statistics. Without parameters,
    this will return the most recent trading session by default.

    Parameters
    ----------
    start : string, int, date, datetime, Timestamp
        The beginning of the date range.
    end : string, int, date, datetime, Timestamp
        The end of the date range.

    Reference: https://www.iextrading.com/developer/docs/#historical-daily

    :return: DataFrame
    """
    from pandas_datareader.iex.stats import DailySummaryReader

    return DailySummaryReader(*args, **kwargs).read()


def get_summary_iex(*args, **kwargs):
    """
    Returns an aggregated monthly summary of market volume and a variety of
    related metrics for trades by lot size, security market cap, and venue.
    In the absence of parameters, this will return month-to-date statistics.
    For ranges spanning multiple months, this will return one row per month.

    start : string, int, date, datetime, Timestamp
        A datetime object - the beginning of the date range.
    end : string, int, date, datetime, Timestamp
        A datetime object - the end of the date range.

    Returns
    -------
    DataFrame
    """
    from pandas_datareader.iex.stats import MonthlySummaryReader

    return MonthlySummaryReader(*args, **kwargs).read()


def get_records_iex(*args, **kwargs):
    """
    Returns the record value, record date, recent value, and 30-day average for
    market volume, # of symbols traded, # of routed trades and notional value.
    This function accepts no additional parameters.

    Reference: https://www.iextrading.com/developer/docs/#records

    :return: DataFrame
    """
    from pandas_datareader.iex.stats import RecordsReader

    return RecordsReader(*args, **kwargs).read()


def get_recent_iex(*args, **kwargs):
    """
    Returns market volume and trade routing statistics for recent sessions.
    Also reports IEX's relative market share, lit share volume and a boolean
    halfday indicator.

    Reference: https://www.iextrading.com/developer/docs/#recent

    Returns
    -------
    DataFrame
    """
    from pandas_datareader.iex.stats import RecentReader

    return RecentReader(*args, **kwargs).read()


def get_iex_symbols(*args, **kwargs):
    """
    Returns a list of all equity symbols available for trading on IEX. Accepts
    no additional parameters.

    Reference: https://www.iextrading.com/developer/docs/#symbols

    :return: DataFrame
    """
    from pandas_datareader.iex.ref import SymbolsReader

    return SymbolsReader(*args, **kwargs).read()


def get_iex_book(*args, **kwargs):
    """
    Returns an array of dictionaries with depth of book data from IEX for up to
    10 securities at a time. Returns a dictionary of the bid and ask books.

    Parameters
    ----------
    symbols : str, List[str]
        A string or list of strings of valid tickers
    service : str
        One of:

        - 'book': Live depth of book data
        - 'op-halt-status': Checks to see if the exchange has instituted a halt
        - 'security-event': Denotes individual security related event
        - 'ssr-status': Short Sale Price Test restrictions, per reg 201 of SHO
        - 'system-event': Relays current feed status (i.e. market open)
        - 'trades': Retrieves recent executions, trade size/price and flags
        - 'trade-breaks': Lists execution breaks for the current trading session
        - 'trading-status': Returns status and cause codes for securities

    Returns
    -------
    DataFrame
    """
    return IEXDeep(*args, **kwargs).read()


@deprecate_kwarg("access_key", "api_key")
def DataReader(
    name,
    data_source=None,
    start=None,
    end=None,
    retry_count=3,
    pause=0.1,
    session=None,
    api_key=None,
):
    """
    Imports data from a number of online sources.

    Currently supports Google Finance, St. Louis FED (FRED),
    and Kenneth French's data library, among others.

    Parameters
    ----------
    name : str or list of strs
        the name of the dataset. Some data sources (IEX, fred) will
        accept a list of names.
    data_source: {str, None}
        the data source ("iex", "fred", "ff")
    start : string, int, date, datetime, Timestamp
        left boundary for range (defaults to 1/1/2010)
    end : string, int, date, datetime, Timestamp
        right boundary for range (defaults to today)
    retry_count : {int, 3}
        Number of times to retry query request.
    pause : {numeric, 0.001}
        Time, in seconds, to pause between consecutive queries of chunks. If
        single value given for symbol, represents the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used
    api_key : (str, None)
        Optional parameter to specify an API key for certain data sources.

    Examples
    ----------
    # Data from Google Finance
    aapl = DataReader("AAPL", "iex")

    # Price and volume data from IEX
    tops = DataReader(["GS", "AAPL"], "iex-tops")
    # Top of book executions from IEX
    gs = DataReader("GS", "iex-last")
    # Real-time depth of book data from IEX
    gs = DataReader("GS", "iex-book")

    # Data from FRED
    vix = DataReader("VIXCLS", "fred")

    # Data from Fama/French
    ff = DataReader("F-F_Research_Data_Factors", "famafrench")
    ff = DataReader("F-F_Research_Data_Factors_weekly", "famafrench")
    ff = DataReader("6_Portfolios_2x3", "famafrench")
    ff = DataReader("F-F_ST_Reversal_Factor", "famafrench")
    """
    expected_source = [
        "yahoo",
        "iex",
        "iex-tops",
        "iex-last",
        "iex-last",
        "bankofcanada",
        "stooq",
        "iex-book",
        "enigma",
        "fred",
        "famafrench",
        "oecd",
        "eurostat",
        "nasdaq",
        "quandl",
        "moex",
        "tiingo",
        "yahoo-actions",
        "yahoo-dividends",
        "av-forex",
        "av-forex-daily",
        "av-daily",
        "av-daily-adjusted",
        "av-weekly",
        "av-weekly-adjusted",
        "av-monthly",
        "av-monthly-adjusted",
        "av-intraday",
        "econdb",
        "naver",
    ]

    if data_source not in expected_source:
        msg = "data_source=%r is not implemented" % data_source
        raise NotImplementedError(msg)

    if data_source == "yahoo":
        return YahooDailyReader(
            symbols=name,
            start=start,
            end=end,
            adjust_price=False,
            chunksize=25,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()

    elif data_source == "iex":
        return IEXDailyReader(
            symbols=name,
            start=start,
            end=end,
            chunksize=25,
            api_key=api_key,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()

    elif data_source == "iex-tops":
        return IEXTops(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()

    elif data_source == "iex-last":
        return IEXLasts(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()

    elif data_source == "bankofcanada":
        return BankOfCanadaReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()

    elif data_source == "stooq":
        return StooqDailyReader(
            symbols=name,
            start=start,
            end=end,
            chunksize=25,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()

    elif data_source == "iex-book":
        return IEXDeep(
            symbols=name,
            service="book",
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()

    elif data_source == "enigma":
        return EnigmaReader(dataset_id=name, api_key=api_key).read()

    elif data_source == "fred":
        return FredReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()

    elif data_source == "famafrench":
        return FamaFrenchReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()

    elif data_source == "oecd":
        return OECDReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()
    elif data_source == "eurostat":
        return EurostatReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()
    elif data_source == "nasdaq":
        if name != "symbols":
            raise ValueError(
                "Only the string 'symbols' is supported for " "Nasdaq, not %r" % (name,)
            )
        return get_nasdaq_symbols(retry_count=retry_count, pause=pause)

    elif data_source == "quandl":
        return QuandlReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            api_key=api_key,
        ).read()
    elif data_source == "moex":
        return MoexReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()
    elif data_source == "tiingo":
        return TiingoDailyReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            api_key=api_key,
        ).read()

    elif data_source == "yahoo-actions":
        return YahooActionReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()

    elif data_source == "yahoo-dividends":
        return YahooDivReader(
            symbols=name,
            start=start,
            end=end,
            adjust_price=False,
            chunksize=25,
            retry_count=retry_count,
            pause=pause,
            session=session,
            interval="d",
        ).read()

    elif data_source == "av-forex":
        return AVForexReader(
            symbols=name,
            retry_count=retry_count,
            pause=pause,
            session=session,
            api_key=api_key,
        ).read()

    elif data_source == "av-forex-daily":
        return AVTimeSeriesReader(
            symbols=name,
            function="FX_DAILY",
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            api_key=api_key,
        ).read()

    elif data_source == "av-daily":
        return AVTimeSeriesReader(
            symbols=name,
            function="TIME_SERIES_DAILY",
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            api_key=api_key,
        ).read()

    elif data_source == "av-daily-adjusted":
        return AVTimeSeriesReader(
            symbols=name,
            function="TIME_SERIES_DAILY_ADJUSTED",
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            api_key=api_key,
        ).read()

    elif data_source == "av-weekly":
        return AVTimeSeriesReader(
            symbols=name,
            function="TIME_SERIES_WEEKLY",
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            api_key=api_key,
        ).read()

    elif data_source == "av-weekly-adjusted":
        return AVTimeSeriesReader(
            symbols=name,
            function="TIME_SERIES_WEEKLY_ADJUSTED",
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            api_key=api_key,
        ).read()

    elif data_source == "av-monthly":
        return AVTimeSeriesReader(
            symbols=name,
            function="TIME_SERIES_MONTHLY",
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            api_key=api_key,
        ).read()

    elif data_source == "av-monthly-adjusted":
        return AVTimeSeriesReader(
            symbols=name,
            function="TIME_SERIES_MONTHLY_ADJUSTED",
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            api_key=api_key,
        ).read()

    elif data_source == "av-intraday":
        return AVTimeSeriesReader(
            symbols=name,
            function="TIME_SERIES_INTRADAY",
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            api_key=api_key,
        ).read()

    elif data_source == "econdb":
        return EcondbReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()

    elif data_source == "naver":
        return NaverDailyReader(
            symbols=name,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        ).read()

    else:
        msg = "data_source=%r is not implemented" % data_source
        raise NotImplementedError(msg)


def Options(symbol, data_source=None, session=None):
    if data_source is None:
        warnings.warn(
            "Options(symbol) is deprecated, use Options(symbol,"
            " data_source) instead",
            FutureWarning,
            stacklevel=2,
        )
        data_source = "yahoo"
    if data_source == "yahoo":
        raise ImmediateDeprecationError(DEP_ERROR_MSG.format("Yahoo Options"))
        return YahooOptions(symbol, session=session)
    else:
        raise NotImplementedError("currently only yahoo supported")



import datetime as dt

from pandas_datareader.av import AlphaVantage


class AVTimeSeriesReader(AlphaVantage):
    """
    Returns DataFrame of the Alpha Vantage Stock Time Series endpoints

    .. versionadded:: 0.7.0

    Parameters
    ----------
    symbols : string
        Single stock symbol (ticker)
    start : string, int, date, datetime, Timestamp
        Starting date. Parses many different kind of date
        representations (e.g., 'JAN-01-2010', '1/1/10', 'Jan, 1, 1980'). Defaults to
        20 years before current date.
    end : string, int, date, datetime, Timestamp
        Ending date
    retry_count : int, default 3
        Number of times to retry query request.
    pause : int, default 0.1
        Time, in seconds, to pause between consecutive queries of chunks. If
        single value given for symbol, represents the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used
    api_key : str, optional
        AlphaVantage API key . If not provided the environmental variable
        ALPHAVANTAGE_API_KEY is read. The API key is *required*.
    """

    _FUNC_TO_DATA_KEY = {
        "TIME_SERIES_DAILY": "Time Series (Daily)",
        "TIME_SERIES_DAILY_ADJUSTED": "Time Series (Daily)",
        "TIME_SERIES_WEEKLY": "Weekly Time Series",
        "TIME_SERIES_WEEKLY_ADJUSTED": "Weekly Adjusted Time Series",
        "TIME_SERIES_MONTHLY": "Monthly Time Series",
        "TIME_SERIES_MONTHLY_ADJUSTED": "Monthly Adjusted Time Series",
        "TIME_SERIES_INTRADAY": "Time Series (1min)",
        "FX_DAILY": "Time Series FX (Daily)",
    }

    def __init__(
        self,
        symbols=None,
        function="TIME_SERIES_DAILY",
        start=None,
        end=None,
        retry_count=3,
        pause=0.1,
        session=None,
        chunksize=25,
        api_key=None,
    ):
        self._func = function
        super(AVTimeSeriesReader, self).__init__(
            symbols=symbols,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            api_key=api_key,
        )

    @property
    def default_start_date(self):
        d_days = 3 if self.intraday else 365 * 20
        return dt.datetime.today() - dt.timedelta(days=d_days)

    @property
    def function(self):
        return self._func

    @property
    def intraday(self):
        return True if self.function == "TIME_SERIES_INTRADAY" else False

    @property
    def forex(self):
        return True if self.function == "FX_DAILY" else False

    @property
    def output_size(self):
        """Used to limit the size of the Alpha Vantage query when
        possible.
        """
        delta = dt.datetime.now() - self.start
        return "compact" if delta.days < 80 and not self.intraday else "full"

    @property
    def data_key(self):
        return self._FUNC_TO_DATA_KEY[self.function]

    @property
    def params(self):
        p = {
            "function": self.function,
            "apikey": self.api_key,
            "outputsize": self.output_size,
        }
        if self.intraday:
            p.update({"interval": "1min"})
        if self.forex:
            p.update({"from_symbol": self.symbols.split("/")[0]})
            p.update({"to_symbol": self.symbols.split("/")[1]})
        else:
            p.update({"symbol": self.symbols})
        return p

    def _read_lines(self, out):
        data = super(AVTimeSeriesReader, self)._read_lines(out)
        # reverse since alphavantage returns descending by date
        data = data[::-1]
        start_str = self.start.strftime("%Y-%m-%d")
        end_str = self.end.strftime("%Y-%m-%d")
        data = data.loc[start_str:end_str]
        if data.empty:
            raise ValueError("Please input a valid date range")
        else:
            for column in data.columns:
                if column == "volume":
                    data[column] = data[column].astype("int64")
                else:
                    data[column] = data[column].astype("float64")
        return data



import json
import time
import warnings

from pandas import DataFrame, Series, concat, to_datetime

from pandas_datareader._utils import RemoteDataError, SymbolWarning
from pandas_datareader.compat import string_types
from pandas_datareader.yahoo.daily import YahooDailyReader


class YahooFXReader(YahooDailyReader):
    """
    Returns DataFrame of historical prices for currencies

    Parameters
    ----------
    symbols : string, array-like object (list, tuple, Series), or DataFrame
        Single stock symbol (ticker), array-like object of symbols or
        DataFrame with index containing stock symbols.
    start : string, int, date, datetime, Timestamp
        Starting date, timestamp. Parses many different kind of date
        representations (e.g., 'JAN-01-2010', '1/1/10', 'Jan, 1, 1980').
        Defaults to '1/1/2010'.
    end : string, int, date, datetime, Timestamp
        Ending date, timestamp. Same format as starting date. Defaults to today.
    retry_count : int, default 3
        Number of times to retry query request.
    pause : int, default 0.1
        Time, in seconds, to pause between consecutive queries of chunks. If
        single value given for symbol, represents the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used
    chunksize : int, default 25 (NOT IMPLEMENTED)
        Number of symbols to download consecutively before intiating pause.
    interval : string, default '1d'
        Valid values are '1d', '5d', '1mo', '3mo', '6mo', '1y', '2y', '5y',
            '10y', 'ytd', 'max'
    """

    def _get_params(self, symbol):
        unix_start = int(time.mktime(self.start.timetuple()))
        day_end = self.end.replace(hour=23, minute=59, second=59)
        unix_end = int(time.mktime(day_end.timetuple()))

        params = {
            "symbol": symbol + "=X",
            "period1": unix_start,
            "period2": unix_end,
            "interval": self.interval,  # deal with this
            "includePrePost": "true",
            "events": "div|split|earn",
            "corsDomain": "finance.yahoo.com",
        }
        return params

    def read(self):
        """Read data"""
        try:
            # If a single symbol, (e.g., 'GOOG')
            if isinstance(self.symbols, (string_types, int)):
                df = self._read_one_data(self.symbols)

            # Or multiple symbols, (e.g., ['GOOG', 'AAPL', 'MSFT'])
            elif isinstance(self.symbols, DataFrame):
                df = self._dl_mult_symbols(self.symbols.index)
            else:
                df = self._dl_mult_symbols(self.symbols)

            if "Date" in df:
                df = df.set_index("Date")

            if "Volume" in df:
                df = df.drop("Volume", axis=1)

            return df.sort_index().dropna(how="all")
        finally:
            self.close()

    def _read_one_data(self, symbol):
        """read one data from specified URL"""
        url = "https://query1.finance.yahoo.com/v8/finance/chart/{}=X".format(symbol)
        params = self._get_params(symbol)

        resp = self._get_response(url, params=params)
        jsn = json.loads(resp.text)

        data = jsn["chart"]["result"][0]
        df = DataFrame(data["indicators"]["quote"][0])
        df.insert(0, "date", to_datetime(Series(data["timestamp"]), unit="s").dt.date)
        df.columns = map(str.capitalize, df.columns)
        return df

    def _dl_mult_symbols(self, symbols):
        stocks = {}
        failed = []
        passed = []
        for sym in symbols:
            try:
                df = self._read_one_data(sym)
                df["PairCode"] = sym
                stocks[sym] = df
                passed.append(sym)
            except IOError:
                msg = "Failed to read symbol: {0!r}, replacing with NaN."
                warnings.warn(msg.format(sym), SymbolWarning)
                failed.append(sym)

        if len(passed) == 0:
            msg = "No data fetched using {0!r}"
            raise RemoteDataError(msg.format(self.__class__.__name__))
        else:
            return concat(stocks).set_index(["PairCode", "Date"])



from datetime import date, datetime, timedelta
import os

from pandas import DataFrame, MultiIndex
import pytest

from pandas_datareader import data as web
from pandas_datareader.iex.daily import IEXDailyReader


@pytest.mark.skipif(
    os.getenv("IEX_SANDBOX") != "enable", reason="All tests must be run in sandbox mode"
)
class TestIEXDaily(object):
    @classmethod
    def setup_class(cls):
        pytest.importorskip("lxml")

    @property
    def start(self):
        return datetime(2015, 2, 9)

    @property
    def end(self):
        return datetime(2017, 5, 24)

    def test_iex_bad_symbol(self):
        with pytest.raises(Exception):
            web.DataReader("BADTICKER", "iex,", self.start, self.end)

    def test_iex_bad_symbol_list(self):
        with pytest.raises(Exception):
            web.DataReader(["AAPL", "BADTICKER"], "iex", self.start, self.end)

    def test_daily_invalid_date(self):
        start = datetime(2000, 1, 5)
        end = datetime(2017, 5, 24)
        with pytest.raises(Exception):
            web.DataReader(["AAPL", "TSLA"], "iex", start, end)

    def test_single_symbol(self):
        df = web.DataReader("AAPL", "iex", self.start, self.end)
        assert list(df) == ["open", "high", "low", "close", "volume"]
        assert len(df) == 476

    def test_multiple_symbols(self):
        syms = ["AAPL", "MSFT", "TSLA"]
        df = web.DataReader(syms, "iex", self.start, self.end)
        assert sorted(list(df.columns.levels[1])) == syms
        for sym in syms:
            assert len(df.xs(sym, level="Symbols", axis=1) == 578)

    def test_multiple_symbols_2(self):
        syms = ["AAPL", "MSFT", "TSLA"]
        good_start = datetime(2017, 2, 9)
        good_end = datetime(2017, 5, 24)
        df = web.DataReader(syms, "iex", good_start, good_end)
        assert isinstance(df, DataFrame)
        assert isinstance(df.columns, MultiIndex)
        assert len(df.columns.levels[1]) == 3
        assert sorted(list(df.columns.levels[1])) == syms

        a = df.xs("AAPL", axis=1, level="Symbols")
        t = df.xs("TSLA", axis=1, level="Symbols")

        assert len(a) == 73
        assert len(t) == 73

    def test_range_string_from_date(self):
        syms = ["AAPL"]

        assert (
            IEXDailyReader(
                symbols=syms, start=date.today() - timedelta(days=5), end=date.today()
            )._range_string_from_date()
            == "5d"
        )
        assert (
            IEXDailyReader(
                symbols=syms, start=date.today() - timedelta(days=27), end=date.today()
            )._range_string_from_date()
            == "1m"
        )
        assert (
            IEXDailyReader(
                symbols=syms, start=date.today() - timedelta(days=83), end=date.today()
            )._range_string_from_date()
            == "3m"
        )
        assert (
            IEXDailyReader(
                symbols=syms, start=date.today() - timedelta(days=167), end=date.today()
            )._range_string_from_date()
            == "6m"
        )
        assert (
            IEXDailyReader(
                symbols=syms, start=date.today() - timedelta(days=170), end=date.today()
            )._range_string_from_date()
            == "1y"
        )
        assert (
            IEXDailyReader(
                symbols=syms, start=date.today() - timedelta(days=365), end=date.today()
            )._range_string_from_date()
            == "1y"
        )
        assert (
            IEXDailyReader(
                symbols=syms, start=date.today() - timedelta(days=730), end=date.today()
            )._range_string_from_date()
            == "2y"
        )
        assert (
            IEXDailyReader(
                symbols=syms,
                start=date.today() - timedelta(days=1826),
                end=date.today(),
            )._range_string_from_date()
            == "5y"
        )



from collections import OrderedDict
import json

from pandas import DataFrame

from pandas_datareader.base import _BaseReader
from pandas_datareader.compat import string_types
from pandas_datareader.yahoo.headers import DEFAULT_HEADERS

_DEFAULT_PARAMS = {
    "lang": "en-US",
    "corsDomain": "finance.yahoo.com",
    ".tsrc": "finance",
}


class YahooQuotesReader(_BaseReader):

    """Get current yahoo quote"""

    def __init__(
        self,
        symbols=None,
        start=None,
        end=None,
        retry_count=3,
        pause=0.1,
        session=None,
    ):
        super().__init__(
            symbols=symbols,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        )
        if session is not None:
            self.headers = session.headers
        else:
            self.headers = DEFAULT_HEADERS

    @property
    def url(self):
        return "https://query1.finance.yahoo.com/v7/finance/quote"

    def read(self):
        if isinstance(self.symbols, string_types):
            return self._read_one_data(self.url, self.params(self.symbols))
        else:
            data = OrderedDict()
            for symbol in self.symbols:
                data[symbol] = self._read_one_data(self.url, self.params(symbol)).loc[
                    symbol
                ]
            return DataFrame.from_dict(data, orient="index")

    def params(self, symbol):
        """Parameters to use in API calls"""
        # Construct the code request string.
        params = {"symbols": symbol}
        params.update(_DEFAULT_PARAMS)
        return params

    def _read_lines(self, out):
        data = json.loads(out.read())["quoteResponse"]["result"][0]
        idx = data.pop("symbol")
        data["price"] = data["regularMarketPrice"]
        return DataFrame(data, index=[idx])



import os

import pandas as pd

from pandas_datareader.base import _BaseReader


def get_tiingo_symbols():
    """
    Get the set of stock symbols supported by Tiingo

    Returns
    -------
    symbols : DataFrame
        DataFrame with symbols (ticker), exchange, asset type, currency and
        start and end dates

    Notes
    -----
    Reads https://apimedia.tiingo.com/docs/tiingo/daily/supported_tickers.zip
    """
    url = "https://apimedia.tiingo.com/docs/tiingo/daily/supported_tickers.zip"
    return pd.read_csv(url)


class TiingoIEXHistoricalReader(_BaseReader):
    """
    Historical data from Tiingo on equities, ETFs and mutual funds,
    with re-sampling capability. This query is limited to the last
    1,000 bars based in the endDate. So the startDate is moved if
    it goes past the limit.

    Parameters
    ----------
    symbols : {str, List[str]}
        String symbol or list of symbols
    start : string, int, date, datetime, Timestamp
        Starting date. Parses many different kind of date
        representations (e.g., 'JAN-01-2010', '1/1/10', 'Jan, 1, 1980'). Defaults to
        20 years before current date.
    end : string, int, date, datetime, Timestamp
        Ending date
    retry_count : int, default 3
        Number of times to retry query request.
    pause : float, default 0.1
        Time, in seconds, of the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used
    freq : {str, None}
    Re-sample frequency. Format is # + (min/hour); e.g. "15min" or "4hour".
    If no value is provided, defaults to 5min. The minimum value is "1min".
    Units in minutes (min) and hours (hour) are accepted.
    api_key : str, optional
        Tiingo API key . If not provided the environmental variable
        TIINGO_API_KEY is read. The API key is *required*.
    """

    def __init__(
        self,
        symbols,
        start=None,
        end=None,
        retry_count=3,
        pause=0.1,
        timeout=30,
        session=None,
        freq=None,
        api_key=None,
    ):
        super().__init__(
            symbols, start, end, retry_count, pause, timeout, session, freq
        )

        if isinstance(self.symbols, str):
            self.symbols = [self.symbols]
        self._symbol = ""
        if api_key is None:
            api_key = os.getenv("TIINGO_API_KEY")
        if not api_key or not isinstance(api_key, str):
            raise ValueError(
                "The tiingo API key must be provided either "
                "through the api_key variable or through the "
                "environmental variable TIINGO_API_KEY."
            )
        self.api_key = api_key
        self._concat_axis = 0

    @property
    def url(self):
        """API URL"""
        _url = "https://api.tiingo.com/iex/{ticker}/prices"
        return _url.format(ticker=self._symbol)

    @property
    def params(self):
        """Parameters to use in API calls"""
        return {
            "startDate": self.start.strftime("%Y-%m-%d"),
            "endDate": self.end.strftime("%Y-%m-%d"),
            "resampleFreq": self.freq,
            "format": "json",
        }

    def _get_crumb(self, *args):
        pass

    def _read_one_data(self, url, params):
        """read one data from specified URL"""
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Token " + self.api_key,
        }
        out = self._get_response(url, params=params, headers=headers).json()
        return self._read_lines(out)

    def _read_lines(self, out):
        df = pd.DataFrame(out)
        df["symbol"] = self._symbol
        df["date"] = pd.to_datetime(df["date"])
        df = df.set_index(["symbol", "date"])
        return df

    def read(self):
        """Read data from connector"""
        dfs = []
        for symbol in self.symbols:
            self._symbol = symbol
            try:
                dfs.append(self._read_one_data(self.url, self.params))
            finally:
                self.close()
        return pd.concat(dfs, self._concat_axis)


class TiingoDailyReader(_BaseReader):
    """
    Historical daily data from Tiingo on equities, ETFs and mutual funds

    Parameters
    ----------
    symbols : {str, List[str]}
        String symbol or list of symbols
    start : string, int, date, datetime, Timestamp
        Starting date, timestamp. Parses many different kind of date
        representations (e.g., 'JAN-01-2010', '1/1/10', 'Jan, 1, 1980').
        Default starting date is 5 years before current date.
    end : string, int, date, datetime, Timestamp
        Ending date, timestamp. Same format as starting date.
    retry_count : int, default 3
        Number of times to retry query request.
    pause : float, default 0.1
        Time, in seconds, of the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used
    freq : {str, None}
        Not used.
    api_key : str, optional
        Tiingo API key . If not provided the environmental variable
        TIINGO_API_KEY is read. The API key is *required*.
    """

    def __init__(
        self,
        symbols,
        start=None,
        end=None,
        retry_count=3,
        pause=0.1,
        timeout=30,
        session=None,
        freq=None,
        api_key=None,
    ):
        super(TiingoDailyReader, self).__init__(
            symbols, start, end, retry_count, pause, timeout, session, freq
        )
        if isinstance(self.symbols, str):
            self.symbols = [self.symbols]
        self._symbol = ""
        if api_key is None:
            api_key = os.getenv("TIINGO_API_KEY")
        if not api_key or not isinstance(api_key, str):
            raise ValueError(
                "The tiingo API key must be provided either "
                "through the api_key variable or through the "
                "environmental variable TIINGO_API_KEY."
            )
        self.api_key = api_key
        self._concat_axis = 0

    @property
    def url(self):
        """API URL"""
        _url = "https://api.tiingo.com/tiingo/daily/{ticker}/prices"
        return _url.format(ticker=self._symbol)

    @property
    def params(self):
        """Parameters to use in API calls"""
        return {
            "startDate": self.start.strftime("%Y-%m-%d"),
            "endDate": self.end.strftime("%Y-%m-%d"),
            "format": "json",
        }

    def _get_crumb(self, *args):
        pass

    def _read_one_data(self, url, params):
        """read one data from specified URL"""
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Token " + self.api_key,
        }
        out = self._get_response(url, params=params, headers=headers).json()
        return self._read_lines(out)

    def _read_lines(self, out):
        df = pd.DataFrame(out)
        df["symbol"] = self._symbol
        df["date"] = pd.to_datetime(df["date"])
        df = df.set_index(["symbol", "date"])
        return df

    def read(self):
        """Read data from connector"""
        dfs = []
        for symbol in self.symbols:
            self._symbol = symbol
            try:
                dfs.append(self._read_one_data(self.url, self.params))
            finally:
                self.close()
        return pd.concat(dfs, self._concat_axis)


class TiingoMetaDataReader(TiingoDailyReader):
    """
    Read metadata about symbols from Tiingo

    Parameters
    ----------
    symbols : {str, List[str]}
        String symbol or list of symbols
    start : string, int, date, datetime, Timestamp
        Not used.
    end : string, int, date, datetime, Timestamp
        Not used.
    retry_count : int, default 3
        Number of times to retry query request.
    pause : float, default 0.1
        Time, in seconds, of the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used
    freq : {str, None}
        Not used.
    api_key : str, optional
        Tiingo API key . If not provided the environmental variable
        TIINGO_API_KEY is read. The API key is *required*.
    """

    def __init__(
        self,
        symbols,
        start=None,
        end=None,
        retry_count=3,
        pause=0.1,
        timeout=30,
        session=None,
        freq=None,
        api_key=None,
    ):
        super(TiingoMetaDataReader, self).__init__(
            symbols, start, end, retry_count, pause, timeout, session, freq, api_key
        )
        self._concat_axis = 1

    @property
    def url(self):
        """API URL"""
        _url = "https://api.tiingo.com/tiingo/daily/{ticker}"
        return _url.format(ticker=self._symbol)

    @property
    def params(self):
        return None

    def _read_lines(self, out):
        s = pd.Series(out)
        s.name = self._symbol
        return s


class TiingoQuoteReader(TiingoDailyReader):
    """
    Read quotes (latest prices) from Tiingo

    Parameters
    ----------
    symbols : {str, List[str]}
        String symbol or list of symbols
    start : string, int, date, datetime, Timestamp
        Not used.
    end : string, int, date, datetime, Timestamp
        Not used.
    retry_count : int, default 3
        Number of times to retry query request.
    pause : float, default 0.1
        Time, in seconds, of the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used
    freq : {str, None}
        Not used.
    api_key : str, optional
        Tiingo API key . If not provided the environmental variable
        TIINGO_API_KEY is read. The API key is *required*.

    Notes
    -----
    This is a special case of the daily reader which automatically selected
    the latest data available for each symbol.
    """

    @property
    def params(self):
        return None



from pandas import DataFrame
import pytest

from pandas_datareader.data import DataReader

pytestmark = pytest.mark.stable


class TestDataReader(object):
    def test_read_iex(self):
        gs = DataReader("GS", "iex-last")
        assert isinstance(gs, DataFrame)

    def test_read_fred(self):
        vix = DataReader("VIXCLS", "fred")
        assert isinstance(vix, DataFrame)

    def test_not_implemented(self):
        with pytest.raises(NotImplementedError):
            DataReader("NA", "NA")



import numpy as np
import pandas as pd
from pandas import testing as tm
import pytest

from pandas_datareader import data as web

pytestmark = pytest.mark.stable


class TestEurostat(object):
    def test_get_ert_h_eur_a(self):
        # Former euro area national currencies vs. euro/ECU
        # annual data (ert_h_eur_a)
        df = web.DataReader(
            "ert_h_eur_a",
            "eurostat",
            start=pd.Timestamp("2009-01-01"),
            end=pd.Timestamp("2010-01-01"),
        )
        assert isinstance(df, pd.DataFrame)

        currencies = ["Italian lira", "Lithuanian litas"]
        # cols = [(currency, "Average") for currency in currencies]
        df_currency = df[currencies]
        df_currency = df_currency.xs("Average", axis=1, level=1)
        df_currency.columns = df_currency.columns.droplevel(1)

        exp_col = pd.MultiIndex.from_product(
            [currencies, ["Annual"]], names=["CURRENCY", "FREQ"]
        )
        exp_idx = pd.DatetimeIndex(["2009-01-01", "2010-01-01"], name="TIME_PERIOD")
        values = np.array([[1936.27, 3.4528], [1936.27, 3.4528]])
        expected = pd.DataFrame(values, index=exp_idx, columns=exp_col)
        tm.assert_frame_equal(df_currency, expected)

    def test_get_sts_cobp_a(self):
        # Building permits - annual data (2010 = 100)
        df = web.DataReader(
            "sts_cobp_a",
            "eurostat",
            start=pd.Timestamp("2000-01-01"),
            end=pd.Timestamp("2013-01-01"),
        )

        idx = pd.date_range("2000-01-01", "2013-01-01", freq="AS", name="TIME_PERIOD")
        ne_name = (
            "Building permits - m2 of useful floor area",
            "Index, 2010=100",
            "Unadjusted data (i.e. neither seasonally adjusted nor "
            "calendar adjusted data)",
            "Non-residential buildings, except office buildings",
            "Netherlands",
            "Annual",
        )
        ne_values = [
            200.0,
            186.5,
            127.3,
            130.7,
            143.3,
            147.8,
            176.7,
            227.4,
            199.4,
            128.5,
            100.0,
            113.8,
            89.3,
            77.6,
        ]
        ne = pd.Series(ne_values, name=ne_name, index=idx)

        uk_name = (
            "Building permits - m2 of useful floor area",
            "Index, 2010=100",
            "Unadjusted data (i.e. neither seasonally adjusted nor "
            "calendar adjusted data)",
            "Non-residential buildings, except office buildings",
            "United Kingdom",
            "Annual",
        )
        uk_values = [
            112.5,
            113.3,
            110.2,
            112.1,
            119.1,
            112.7,
            113.1,
            121.8,
            114.0,
            105.9,
            100.0,
            98.6,
            103.7,
            81.3,
        ]
        uk = pd.Series(uk_values, name=uk_name, index=idx)

        for expected in [ne, uk]:
            expected.index.freq = None
            result = df[expected.name]
            tm.assert_series_equal(result, expected)

    def test_get_nrg_pc_202(self):
        # see gh-149

        df = web.DataReader(
            "nrg_pc_202",
            "eurostat",
            start=pd.Timestamp("2010-01-01"),
            end=pd.Timestamp("2013-01-01"),
        )

        name = (
            "Euro",
            "Gigajoule (gross calorific value - GCV)",
            "Natural gas",
            "All taxes and levies included",
            "Band D1 : Consumption < 20 GJ",
            "Denmark",
            "Semi-annual",
        )

        exp_index = pd.Index(
            [
                "2013-B2",
                "2013-B1",
                "2012-B2",
                "2012-B1",
                "2011-B2",
                "2011-B1",
                "2010-B2",
                "2010-B1",
            ],
            name="TIME_PERIOD",
        )
        exp = pd.Series(
            [27.1403, 27.5854, 26.5285, 27.2187, 28.5862, 28.6448, 26.8147, 26.4979],
            index=exp_index,
            name=name,
        )

        tm.assert_series_equal(df[name], exp)

    def test_get_prc_hicp_manr_exceeds_limit(self):
        # see gh-149
        msg = "Query size exceeds maximum limit"
        with pytest.raises(ValueError, match=msg):
            web.DataReader(
                "prc_hicp_manr",
                "eurostat",
                start=pd.Timestamp("2000-01-01"),
                end=pd.Timestamp("2013-01-01"),
            )



# pylint: disable-msg=E1101,W0613,W0603

import os

import numpy as np
import pandas as pd
from pandas import testing as tm
import pytest

from pandas_datareader.io.sdmx import _read_sdmx_dsd, read_sdmx

pytestmark = pytest.mark.stable


@pytest.fixture
def dirpath(datapath):
    return datapath("io", "data")


def test_tourism(dirpath):
    # Eurostat
    # Employed doctorate holders in non managerial and non professional
    # occupations by fields of science (%)
    dsd = _read_sdmx_dsd(os.path.join(dirpath, "sdmx", "DSD_cdh_e_fos.xml"))
    df = read_sdmx(os.path.join(dirpath, "sdmx", "cdh_e_fos.xml"), dsd=dsd)

    assert isinstance(df, pd.DataFrame)
    assert df.shape == (2, 336)

    df = df["Percentage"]["Total"]["Natural sciences"]
    df = df[["Norway", "Poland", "Portugal", "Russia"]]

    exp_col = pd.MultiIndex.from_product(
        [["Norway", "Poland", "Portugal", "Russia"], ["Annual"]], names=["GEO", "FREQ"]
    )
    exp_idx = pd.DatetimeIndex(["2009", "2006"], name="TIME_PERIOD")

    values = np.array([[20.38, 25.1, 27.77, 38.1], [25.49, np.nan, 39.05, np.nan]])
    expected = pd.DataFrame(values, index=exp_idx, columns=exp_col)
    tm.assert_frame_equal(df, expected)



from __future__ import unicode_literals

import pandas as pd

from pandas_datareader.base import _BaseReader
from pandas_datareader.compat import string_types
from pandas_datareader.io.sdmx import _read_sdmx_dsd, read_sdmx


class EurostatReader(_BaseReader):
    """Get data for the given name from Eurostat."""

    _URL = "http://ec.europa.eu/eurostat/SDMX/diss-web/rest"

    @property
    def url(self):
        """API URL"""
        if not isinstance(self.symbols, string_types):
            raise ValueError("data name must be string")

        q = "{0}/data/{1}/?startperiod={2}&endperiod={3}"
        return q.format(self._URL, self.symbols, self.start.year, self.end.year)

    @property
    def dsd_url(self):
        """API DSD URL"""
        if not isinstance(self.symbols, string_types):
            raise ValueError("data name must be string")

        return "{0}/datastructure/ESTAT/DSD_{1}".format(self._URL, self.symbols)

    def _read_one_data(self, url, params):
        resp_dsd = self._get_response(self.dsd_url)
        dsd = _read_sdmx_dsd(resp_dsd.content)

        resp = self._get_response(url)
        data = read_sdmx(resp.content, dsd=dsd)

        try:
            data.index = pd.to_datetime(data.index)
            data = data.sort_index()
        except ValueError:
            pass

        try:
            data = data.truncate(self.start, self.end)
        except TypeError:
            pass

        return data



from pandas import DataFrame, MultiIndex

from pandas_datareader.compat import concat
from pandas_datareader.yahoo.daily import YahooDailyReader


class YahooActionReader(YahooDailyReader):
    """
    Returns DataFrame of historical corporate actions (dividends and stock
    splits) from symbols, over date range, start to end. All dates in the
    resulting DataFrame correspond with dividend and stock split ex-dates.
    """

    def read(self):
        data = super().read()
        actions = {}
        if isinstance(data.columns, MultiIndex):
            data = data.swaplevel(0, 1, axis=1)
            for s in data.columns.levels[0]:
                actions[s] = _get_one_action(data[s])
            return actions
        else:
            return _get_one_action(data)

    @property
    def get_actions(self):
        return True


def _get_one_action(data):
    actions = DataFrame(columns=["action", "value"])

    if "Dividends" in data.columns:
        # Add a label column so we can combine our two DFs
        dividends = DataFrame(data["Dividends"]).dropna()
        dividends["action"] = "DIVIDEND"
        dividends = dividends.rename(columns={"Dividends": "value"})
        actions = concat([actions, dividends], sort=True)
        actions = actions.sort_index(ascending=False)

    if "Splits" in data.columns:
        # Add a label column so we can combine our two DFs
        splits = DataFrame(data["Splits"]).dropna()
        splits["action"] = "SPLIT"
        splits = splits.rename(columns={"Splits": "value"})
        actions = concat([actions, splits], sort=True)
        actions = actions.sort_index(ascending=False)

    return actions


class YahooDivReader(YahooActionReader):
    def read(self):
        data = super(YahooDivReader, self).read()
        return data[data["action"] == "DIVIDEND"]


class YahooSplitReader(YahooActionReader):
    def read(self):
        data = super(YahooSplitReader, self).read()
        return data[data["action"] == "SPLIT"]



from __future__ import unicode_literals

import collections
from io import BytesIO
import time
import zipfile

import pandas as pd

from pandas_datareader.compat import HTTPError, str_to_bytes
from pandas_datareader.io.util import _read_content

_STRUCTURE = "{http://www.sdmx.org/resources/sdmxml/schemas/v2_1/structure}"
_MESSAGE = "{http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message}"
_GENERIC = "{http://www.sdmx.org/resources/sdmxml/schemas/v2_1/data/generic}"
_COMMON = "{http://www.sdmx.org/resources/sdmxml/schemas/v2_1/common}"
_XML = "{http://www.w3.org/XML/1998/namespace}"

_DATASET = _MESSAGE + "DataSet"
_SERIES = _GENERIC + "Series"
_SERIES_KEY = _GENERIC + "SeriesKey"
_OBSERVATION = _GENERIC + "Obs"
_VALUE = _GENERIC + "Value"
_OBSDIMENSION = _GENERIC + "ObsDimension"
_OBSVALUE = _GENERIC + "ObsValue"
_CODE = _STRUCTURE + "Code"
_TIMEDIMENSION = _STRUCTURE + "TimeDimension"


def read_sdmx(path_or_buf, dtype="float64", dsd=None):
    """
    Convert a SDMX-XML string to pandas object

    Parameters
    ----------
    path_or_buf : a valid SDMX-XML string or file-like
        https://webgate.ec.europa.eu/fpfis/mwikis/sdmx/index.php/Main_Page
    dtype : str
        dtype to coerce values
    dsd : dict
        parsed DSD dict corresponding to the SDMX-XML data

    Returns
    -------
    results : Series, DataFrame, or dictionaly of Series or DataFrame.
    """

    xdata = _read_content(path_or_buf)

    from xml.etree import ElementTree as ET

    root = ET.fromstring(xdata)

    try:
        structure = _get_child(root, _MESSAGE + "Structure")
    except ValueError:
        # get zipped path
        result = list(root.iter(_COMMON + "Text"))[1].text
        if not result.startswith("http"):
            raise ValueError(result)

        for _ in range(60):
            # wait zipped data is prepared
            try:
                data = _read_zipped_sdmx(result)
                return read_sdmx(data, dtype=dtype, dsd=dsd)
            except HTTPError:
                time.sleep(1)
                continue

        msg = (
            "Unable to download zipped data within 60 secs, "
            "please download it manually from: {0}"
        )
        raise ValueError(msg.format(result))

    idx_name = structure.get("dimensionAtObservation")
    dataset = _get_child(root, _DATASET)

    keys = []
    obss = []

    for series in dataset.iter(_SERIES):
        key = _parse_series_key(series)
        obs = _parse_observations(series.iter(_OBSERVATION))
        keys.append(key)
        obss.append(obs)

    mcols = _construct_index(keys, dsd=dsd)
    mseries = _construct_series(obss, name=idx_name, dsd=dsd)

    df = pd.DataFrame(mseries, dtype=dtype)
    df = df.T
    df.columns = mcols

    return df


def _construct_series(values, name, dsd=None):

    # ts defines attributes to be handled as times
    times = dsd.ts if dsd is not None else []

    if len(values) < 1:
        raise ValueError("Data contains no 'Series'")
    results = []
    for value in values:

        if name in times:
            tvalue = [v[0] for v in value]
            try:
                idx = pd.DatetimeIndex(tvalue, name=name)
            except ValueError:
                # time may be unsupported format, like '2015-B1'
                idx = pd.Index(tvalue, name=name)
        else:
            idx = pd.Index([v[0] for v in value], name=name)

        results.append(pd.Series([v[1] for v in value], index=idx))
    return results


def _construct_index(keys, dsd=None):

    # code defines a mapping to key's internal code to its representation
    codes = dsd.codes if dsd is not None else {}

    if len(keys) < 1:
        raise ValueError("Data contains no 'Series'")
    names = [t[0] for t in keys[0]]
    values = {}
    # initialize
    for key in keys:
        for name, value in key:
            # apply DSD
            try:
                value = codes[name][value]
            except KeyError:
                pass

            try:
                values[name].append(value)
            except KeyError:
                values[name] = [value]

    midx = pd.MultiIndex.from_arrays([values[name] for name in names], names=names)
    return midx


def _parse_observations(observations):
    results = []
    for observation in observations:
        obsdimension = _get_child(observation, _OBSDIMENSION)
        obsvalue = _get_child(observation, _OBSVALUE)
        results.append((obsdimension.get("value"), obsvalue.get("value")))
    # return list of key/value tuple, eg: [(key, value), ...]
    return results


def _parse_series_key(series):
    serieskey = _get_child(series, _SERIES_KEY)
    key_values = serieskey.iter(_VALUE)
    keys = [(key.get("id"), key.get("value")) for key in key_values]
    # return list of key/value tuple, eg: [(key, value), ...]
    return keys


def _get_child(element, key):
    elements = list(element.iter(key))
    if len(elements) == 1:
        return elements[0]
    elif len(elements) == 0:
        raise ValueError("Element {0} contains " "no {1}".format(element.tag, key))
    else:
        raise ValueError(
            "Element {0} contains " "multiple {1}".format(element.tag, key)
        )


_NAME_EN = ".//{0}Name[@{1}lang='en']".format(_COMMON, _XML)


def _get_english_name(element):
    name = element.find(_NAME_EN).text
    return name


SDMXCode = collections.namedtuple("SDMXCode", ["codes", "ts"])


def _read_sdmx_dsd(path_or_buf):
    """
    Convert a SDMX-XML DSD string to mapping dictionary

    Parameters
    ----------
    filepath_or_buffer : a valid SDMX-XML DSD string or file-like
        https://webgate.ec.europa.eu/fpfis/mwikis/sdmx/index.php/Main_Page

    Returns
    -------
    results : namedtuple (SDMXCode)
    """

    xdata = _read_content(path_or_buf)

    from xml.etree import cElementTree as ET

    root = ET.fromstring(xdata)

    structure = _get_child(root, _MESSAGE + "Structures")
    codes = _get_child(structure, _STRUCTURE + "Codelists")
    # concepts = _get_child(structure, _STRUCTURE + 'Concepts')
    datastructures = _get_child(structure, _STRUCTURE + "DataStructures")

    code_results = {}
    for codelist in codes:
        # codelist_id = codelist.get('id')
        codelist_name = _get_english_name(codelist)
        mapper = {}
        for code in codelist.iter(_CODE):
            code_id = code.get("id")
            name = _get_english_name(code)
            mapper[code_id] = name
        # codeobj = SDMXCode(id=codelist_id, name=codelist_name, mapper=mapper)
        # code_results[codelist_id] = codeobj
        code_results[codelist_name] = mapper

    times = list(datastructures.iter(_TIMEDIMENSION))
    times = [t.get("id") for t in times]

    result = SDMXCode(codes=code_results, ts=times)
    return result


def _read_zipped_sdmx(path_or_buf):
    """Unzipp data contains SDMX-XML"""
    data = _read_content(path_or_buf)

    zp = BytesIO()
    zp.write(str_to_bytes(data))
    f = zipfile.ZipFile(zp)
    files = f.namelist()
    assert len(files) == 1
    return f.open(files[0])



import os

import pandas as pd
import pytest

from pandas_datareader import data as web

TEST_API_KEY = os.getenv("ALPHAVANTAGE_API_KEY")
TEST_API_KEY = None if not TEST_API_KEY else TEST_API_KEY

pytestmark = [
    pytest.mark.requires_api_key,
    pytest.mark.alpha_vantage,
    pytest.mark.skipif(TEST_API_KEY is None, reason="ALPHAVANTAGE_API_KEY not set"),
]


class TestAlphaVantageForex(object):
    @classmethod
    def setup_class(cls):
        pytest.importorskip("lxml")

    @pytest.mark.skipif(TEST_API_KEY is None, reason="ALPHAVANTAGE_API_KEY not set")
    def test_bad_pair_format(self):
        with pytest.raises(ValueError):
            web.DataReader("BAD FORMAT", "av-forex")

    @pytest.mark.skipif(TEST_API_KEY is None, reason="ALPHAVANTAGE_API_KEY not set")
    def test_bad_pairs_format(self):
        with pytest.raises(ValueError):
            web.DataReader(["USD/JPY", "BAD FORMAT"], "av-forex")

    @pytest.mark.skipif(TEST_API_KEY is None, reason="ALPHAVANTAGE_API_KEY not set")
    def test_one_pair(self):
        df = web.DataReader("USD/EUR", "av-forex", retry_count=6, pause=20.5)
        assert isinstance(df, pd.DataFrame)
        assert df.loc["To_Currency Name"][0] == "Euro"
        assert df.loc["Time Zone"][0] == "UTC"

    @pytest.mark.skipif(TEST_API_KEY is None, reason="ALPHAVANTAGE_API_KEY not set")
    def test_multiple_pairs(self):
        pairs = ["USD/JPY", "EUR/JPY"]
        df = web.DataReader(pairs, "av-forex", retry_count=6, pause=20.5)
        assert isinstance(df, pd.DataFrame)
        assert df.columns.equals(pd.Index(pairs))



import numpy as np
import pandas as pd

from pandas_datareader.av import AlphaVantage
from pandas_datareader.exceptions import DEP_ERROR_MSG, ImmediateDeprecationError


class AVQuotesReader(AlphaVantage):
    """
    Returns DataFrame of Alpha Vantage Realtime Stock quotes for a symbol or
    list of symbols.

    Parameters
    ----------
    symbols : string, array-like object (list, tuple, Series), or DataFrame
        Single stock symbol (ticker), array-like object of symbols or
        DataFrame with index containing stock symbols.
    retry_count : int, default 3
        Number of times to retry query request.
    pause : int, default 0.1
        Time, in seconds, to pause between consecutive queries of chunks. If
        single value given for symbol, represents the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used
    """

    def __init__(
        self, symbols=None, retry_count=3, pause=0.1, session=None, api_key=None
    ):
        raise ImmediateDeprecationError(DEP_ERROR_MSG.format("AVQuotesReader"))

        if isinstance(symbols, str):
            syms = [symbols]
        elif isinstance(symbols, list):
            if len(symbols) > 100:
                raise ValueError("Up to 100 symbols at once are allowed.")
            else:
                syms = symbols
        super(AVQuotesReader, self).__init__(
            symbols=syms,
            start=None,
            end=None,
            retry_count=retry_count,
            pause=pause,
            session=session,
            api_key=api_key,
        )

    @property
    def function(self):
        return "BATCH_STOCK_QUOTES"

    @property
    def data_key(self):
        return "Stock Quotes"

    @property
    def params(self):
        return {
            "symbols": ",".join(self.symbols),
            "function": self.function,
            "apikey": self.api_key,
        }

    def _read_lines(self, out):
        result = []
        quotes = out[self.data_key]
        for quote in quotes:
            df = pd.DataFrame(quote, index=[0])
            df.columns = [col[3:] for col in df.columns]
            df.set_index("symbol", inplace=True)
            df["price"] = df["price"].astype("float64")
            try:
                df["volume"] = df["volume"].astype("int64")
            except ValueError:
                df["volume"] = [np.nan * len(self.symbols)]
            result.append(df)
        if len(result) != len(self.symbols):
            raise ValueError("Not all symbols downloaded. Check valid " "ticker(s).")
        else:
            return pd.concat(result)



from datetime import datetime

from pandas_datareader.iex import IEX

# Data provided for free by IEX
# Data is furnished in compliance with the guidelines promulgated in the IEX
# API terms of service and manual
# See https://iextrading.com/api-exhibit-a/ for additional information
# and conditions of use


class Deep(IEX):
    """
    Retrieve order book data from IEX

    Notes
    -----
    Real-time depth of book quotations direct from IEX. Returns aggregated
    size of resting displayed orders at a price and side. Does not indicate
    the size or number of individual orders at any price level. Non-displayed
    orders and non-displayed portions of reserve orders are not counted.

    Also provides last trade price and size information. Routed executions
    are not reported.
    """

    def __init__(
        self,
        symbols=None,
        service=None,
        start=None,
        end=None,
        retry_count=3,
        pause=0.1,
        session=None,
    ):
        if isinstance(symbols, str):
            symbols = symbols.lower()
        else:
            symbols = [s.lower() for s in symbols]

        super(Deep, self).__init__(
            symbols=symbols,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        )
        self.sub = service

    @property
    def service(self):
        """Service endpoint"""
        ss = "/" + self.sub if self.sub is not None else ""
        return "deep{}".format(ss)

    def _read_lines(self, out):
        """
        IEX depth of book data varies and shouldn't always be returned in a DF

        Parameters
        ----------
        out: bytes
            The raw output from an HTTP request

        Returns
        -------
        DataFrame
        """

        # Runs appropriate output functions per the service being accessed.
        fmap = {
            "book": "_pass",
            "op-halt-status": "_convert_tstamp",
            "security-event": "_convert_tstamp",
            "ssr-status": "_convert_tstamp",
            "system-event": "_read_system_event",
            "trades": "_pass",
            "trade-breaks": "_convert_tstamp",
            "trading-status": "_read_trading_status",
            None: "_pass",
        }

        if self.sub in fmap:
            return getattr(self, fmap[self.sub])(out)
        else:
            raise "Invalid service specified: {}.".format(self.sub)

    def _read_system_event(self, out):
        # Map the response code to a string output per the API docs.
        # Per: https://www.iextrading.com/developer/docs/#system-event-message
        smap = {
            "O": "Start of messages",
            "S": "Start of system hours",
            "R": "Start of regular market hours",
            "M": "End of regular market hours",
            "E": "End of system hours",
            "C": "End of messages",
        }
        tid = out["systemEvent"]
        out["eventResponse"] = smap[tid]

        return self._convert_tstamp(out)

    @staticmethod
    def _pass(out):
        return out

    def _read_trading_status(self, out):
        # Reference: https://www.iextrading.com/developer/docs/#trading-status
        smap = {
            "H": "Trading halted across all US equity markets",
            "O": "Trading halt released into an Order Acceptance Period "
            "(IEX-listed securities only)",
            "P": "Trading paused and Order Acceptance Period on IEX "
            "(IEX-listed securities only)",
            "T": "Trading on IEX",
        }
        rmap = {
            # Trading Halt Reasons
            "T1": "Halt News Pending",
            "IPO1": "IPO/New Issue Not Yet Trading",
            "IPOD": "IPO/New Issue Deferred",
            "MCB3": "Market-Wide Circuit Breaker Level 3 - Breached",
            "NA": "Reason Not Available",
            # Order Acceptance Period Reasons
            "T2": "Halt News Dissemination",
            "IPO2": "IPO/New Issue Order Acceptance Period",
            "IPO3": "IPO Pre-Launch Period",
            "MCB1": "Market-Wide Circuit Breaker Level 1 - Breached",
            "MCB2": "Market-Wide Circuit Breaker Level 2 - Breached",
        }
        for ticker, data in out.items():
            if data["status"] in smap:
                data["statusText"] = smap[data["status"]]

            if data["reason"] in rmap:
                data["reasonText"] = rmap[data["reason"]]

            out[ticker] = data

        return self._convert_tstamp(out)

    @staticmethod
    def _convert_tstamp(out):
        # Searches for top-level timestamp attributes or within dictionaries
        if "timestamp" in out:
            # Convert UNIX to datetime object
            f = float(out["timestamp"])
            out["timestamp"] = datetime.fromtimestamp(f / 1000)
        else:
            for ticker, data in out.items():
                if "timestamp" in data:
                    f = float(data["timestamp"])
                    data["timestamp"] = datetime.fromtimestamp(f / 1000)
                    out[ticker] = data

        return out



import datetime
import time
from urllib.parse import urlencode
import warnings

import numpy as np
from pandas import DataFrame, concat, read_csv
import requests

from pandas_datareader._utils import (
    RemoteDataError,
    SymbolWarning,
    _init_session,
    _sanitize_dates,
)
from pandas_datareader.compat import (
    PANDAS_0230,
    StringIO,
    binary_type,
    bytes_to_str,
    string_types,
)


class _BaseReader(object):
    """
    Parameters
    ----------
    symbols : {str, List[str]}
        String symbol of like of symbols
    start : string, int, date, datetime, Timestamp
        Starting date. Parses many different kind of date
        representations (e.g., 'JAN-01-2010', '1/1/10', 'Jan, 1, 1980')
    end : string, int, date, datetime, Timestamp
        Ending date
    retry_count : int, default 3
        Number of times to retry query request.
    pause : float, default 0.1
        Time, in seconds, of the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used.
    freq : {str, None}
        Frequency to use in select readers
    """

    _chunk_size = 1024 * 1024
    _format = "string"

    def __init__(
        self,
        symbols,
        start=None,
        end=None,
        retry_count=3,
        pause=0.1,
        timeout=30,
        session=None,
        freq=None,
    ):

        self.symbols = symbols

        start, end = _sanitize_dates(start or self.default_start_date, end)
        self.start = start
        self.end = end

        if not isinstance(retry_count, int) or retry_count < 0:
            raise ValueError("'retry_count' must be integer larger than 0")
        self.retry_count = retry_count
        self.pause = pause
        self.timeout = timeout
        self.pause_multiplier = 1
        self.session = _init_session(session)
        self.freq = freq
        self.headers = None

    def close(self):
        """Close network session"""
        self.session.close()

    @property
    def default_start_date(self):
        """Default start date for reader. Defaults to 5 years before current date"""
        today = datetime.date.today()
        return today - datetime.timedelta(days=365 * 5)

    @property
    def url(self):
        """API URL"""
        # must be overridden in subclass
        raise NotImplementedError

    @property
    def params(self):
        """Parameters to use in API calls"""
        return None

    def read(self):
        """Read data from connector"""
        try:
            return self._read_one_data(self.url, self.params)
        finally:
            self.close()

    def _read_one_data(self, url, params):
        """read one data from specified URL"""
        if self._format == "string":
            out = self._read_url_as_StringIO(url, params=params)
        elif self._format == "json":
            out = self._get_response(url, params=params).json()
        else:
            raise NotImplementedError(self._format)
        return self._read_lines(out)

    def _read_url_as_StringIO(self, url, params=None):
        """
        Open url (and retry)
        """
        response = self._get_response(url, params=params)
        text = self._sanitize_response(response)
        out = StringIO()
        if len(text) == 0:
            service = self.__class__.__name__
            raise IOError(
                "{} request returned no data; check URL for invalid "
                "inputs: {}".format(service, self.url)
            )
        if isinstance(text, binary_type):
            out.write(bytes_to_str(text))
        else:
            out.write(text)
        out.seek(0)
        return out

    @staticmethod
    def _sanitize_response(response):
        """
        Hook to allow subclasses to clean up response data
        """
        return response.content

    def _get_response(self, url, params=None, headers=None):
        """send raw HTTP request to get requests.Response from the specified url
        Parameters
        ----------
        url : str
            target URL
        params : dict or None
            parameters passed to the URL
        """
        headers = headers or self.headers
        pause = self.pause
        last_response_text = ""
        for _ in range(self.retry_count + 1):
            response = self.session.get(
                url, params=params, headers=headers, timeout=self.timeout
            )
            if response.status_code == requests.codes.ok:
                return response

            if response.encoding:
                last_response_text = response.text.encode(response.encoding)
            time.sleep(pause)

            # Increase time between subsequent requests, per subclass.
            pause *= self.pause_multiplier
            # Get a new breadcrumb if necessary, in case ours is invalidated
            if isinstance(params, list) and "crumb" in params:
                params["crumb"] = self._get_crumb(self.retry_count)

            # If our output error function returns True, exit the loop.
            if self._output_error(response):
                break

        if params is not None and len(params) > 0:
            url = url + "?" + urlencode(params)
        msg = "Unable to read URL: {0}".format(url)
        if last_response_text:
            msg += "\nResponse Text:\n{0}".format(last_response_text)

        raise RemoteDataError(msg)

    def _get_crumb(self, *args):
        """To be implemented by subclass"""
        raise NotImplementedError("Subclass has not implemented method.")

    def _output_error(self, out):
        """If necessary, a service can implement an interpreter for any non-200
         HTTP responses.

        Parameters
        ----------
        out: bytes
            The raw output from an HTTP request

        Returns
        -------
        boolean
        """
        return False

    def _read_lines(self, out):
        rs = read_csv(out, index_col=0, parse_dates=True, na_values=("-", "null"))[::-1]
        # Needed to remove blank space character in header names
        rs.columns = list(map(lambda x: x.strip(), rs.columns.values.tolist()))

        # Yahoo! Finance sometimes does this awesome thing where they
        # return 2 rows for the most recent business day
        if len(rs) > 2 and rs.index[-1] == rs.index[-2]:  # pragma: no cover
            rs = rs[:-1]
        # Get rid of unicode characters in index name.
        try:
            rs.index.name = rs.index.name.decode("unicode_escape").encode(
                "ascii", "ignore"
            )
        except AttributeError:
            # Python 3 string has no decode method.
            rs.index.name = rs.index.name.encode("ascii", "ignore").decode()

        return rs


class _DailyBaseReader(_BaseReader):
    """Base class for Google / Yahoo daily reader"""

    def __init__(
        self,
        symbols=None,
        start=None,
        end=None,
        retry_count=3,
        pause=0.1,
        session=None,
        chunksize=25,
    ):
        super(_DailyBaseReader, self).__init__(
            symbols=symbols,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
        )
        self.chunksize = chunksize

    def _get_params(self, *args, **kwargs):
        raise NotImplementedError

    def read(self):
        """Read data"""
        # If a single symbol, (e.g., 'GOOG')
        if isinstance(self.symbols, (string_types, int)):
            df = self._read_one_data(self.url, params=self._get_params(self.symbols))
        # Or multiple symbols, (e.g., ['GOOG', 'AAPL', 'MSFT'])
        elif isinstance(self.symbols, DataFrame):
            df = self._dl_mult_symbols(self.symbols.index)
        else:
            df = self._dl_mult_symbols(self.symbols)
        return df

    def _dl_mult_symbols(self, symbols):
        stocks = {}
        failed = []
        passed = []
        for sym_group in _in_chunks(symbols, self.chunksize):
            for sym in sym_group:
                try:
                    stocks[sym] = self._read_one_data(self.url, self._get_params(sym))
                    passed.append(sym)
                except (IOError, KeyError):
                    msg = "Failed to read symbol: {0!r}, replacing with NaN."
                    warnings.warn(msg.format(sym), SymbolWarning)
                    failed.append(sym)

        if len(passed) == 0:
            msg = "No data fetched using {0!r}"
            raise RemoteDataError(msg.format(self.__class__.__name__))
        try:
            if len(stocks) > 0 and len(failed) > 0 and len(passed) > 0:
                df_na = stocks[passed[0]].copy()
                df_na[:] = np.nan
                for sym in failed:
                    stocks[sym] = df_na
            if PANDAS_0230:
                result = concat(stocks, sort=True).unstack(level=0)
            else:
                result = concat(stocks).unstack(level=0)
            result.columns.names = ["Attributes", "Symbols"]
            return result
        except AttributeError:
            # cannot construct a panel with just 1D nans indicating no data
            msg = "No data fetched using {0!r}"
            raise RemoteDataError(msg.format(self.__class__.__name__))


def _in_chunks(seq, size):
    """
    Return sequence in 'chunks' of size defined by size
    """
    return (seq[pos : pos + size] for pos in range(0, len(seq), size))


class _OptionBaseReader(_BaseReader):
    def __init__(self, symbol, session=None):
        """Instantiates options_data with a ticker saved as symbol"""
        self.symbol = symbol.upper()
        super(_OptionBaseReader, self).__init__(symbols=symbol, session=session)

    def get_options_data(self, month=None, year=None, expiry=None):
        """
        ***Experimental***
        Gets call/put data for the stock with the expiration data in the
        given month and year
        """
        raise NotImplementedError

    def get_call_data(self, month=None, year=None, expiry=None):
        """
        ***Experimental***
        Gets call/put data for the stock with the expiration data in the
        given month and year
        """
        raise NotImplementedError

    def get_put_data(self, month=None, year=None, expiry=None):
        """
        ***Experimental***
        Gets put data for the stock with the expiration data in the
        given month and year
        """
        raise NotImplementedError

    def get_near_stock_price(
        self, above_below=2, call=True, put=False, month=None, year=None, expiry=None
    ):
        """
        ***Experimental***
        Returns a data frame of options that are near the current stock price.
        """
        raise NotImplementedError

    def get_forward_data(
        self, months, call=True, put=False, near=False, above_below=2
    ):  # pragma: no cover
        """
        ***Experimental***
        Gets either call, put, or both data for months starting in the current
        month and going out in the future a specified amount of time.
        """
        raise NotImplementedError

    def get_all_data(self, call=True, put=True):
        """
        ***Experimental***
        Gets either call, put, or both data for all available months starting
        in the current month.
        """
        raise NotImplementedError



import os
import time

import pandas as pd

from pandas_datareader.base import _BaseReader, string_types
from pandas_datareader.compat import StringIO
from pandas_datareader.exceptions import DEP_ERROR_MSG, ImmediateDeprecationError


class EnigmaReader(_BaseReader):
    """
    Collects current snapshot of Enigma data located at the specified
    data set ID and returns a pandas DataFrame.

    Parameters
    ----------
    dataset_id : str
        Enigma dataset UUID.
    api_key : str, optional
        Enigma API key. If not provided, the environmental variable
        ENIGMA_API_KEY is read.
    retry_count : int, default 5
        Number of times to retry query request.
    pause : float, default 0.1
        Time, in seconds, of the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used.
    base_url : str, optional (defaults to https://public.enigma.com/api)
        Alternative Enigma endpoint to be used.

    Examples
    --------
    Download current snapshot for the following Florida Inspections Dataset:
    https://public.enigma.com/datasets/bedaf052-5fcd-4758-8d27-048ce8746c6a

    >>> import pandas_datareader as pdr
    >>> df = pdr.get_data_enigma('bedaf052-5fcd-4758-8d27-048ce8746c6a')

    In the event that ENIGMA_API_KEY does not exist in your env, the key can
    be supplied as the second argument or as the keyword argument `api_key`

    >>> df = EnigmaReader(dataset_id='bedaf052-5fcd-4758-8d27-048ce8746c6a',
    ...                   api_key='INSERT_API_KEY').read()
    """

    def __init__(
        self,
        dataset_id=None,
        api_key=None,
        retry_count=5,
        pause=0.75,
        session=None,
        base_url="https://public.enigma.com/api",
    ):
        raise ImmediateDeprecationError(DEP_ERROR_MSG.format("Enigma"))

        super(EnigmaReader, self).__init__(
            symbols=[], retry_count=retry_count, pause=pause, session=session
        )
        if api_key is None:
            self._api_key = os.getenv("ENIGMA_API_KEY")
            if self._api_key is None:
                raise ValueError(
                    "Please provide an Enigma API key or set "
                    "the ENIGMA_API_KEY environment variable\n"
                    "If you do not have an API key, you can get "
                    "one here: http://public.enigma.com/signup"
                )
        else:
            self._api_key = api_key

        self._dataset_id = dataset_id
        if not isinstance(self._dataset_id, string_types):
            raise ValueError(
                "The Enigma dataset_id must be a string (ex: "
                "'bedaf052-5fcd-4758-8d27-048ce8746c6a')"
            )

        headers = {
            "Authorization": "Bearer {0}".format(self._api_key),
            "User-Agent": "pandas-datareader",
        }
        self.session.headers.update(headers)
        self._base_url = base_url
        self._retry_count = retry_count
        self._retry_delay = pause

    def read(self):
        """Read data"""
        try:
            return self._read()
        finally:
            self.close()

    def _read(self):
        snapshot_id = self.get_current_snapshot_id(self._dataset_id)
        exported_data = self.get_snapshot_export(snapshot_id)  # TODO: Retry?
        decoded_data = exported_data.decode("utf-8")
        return pd.read_csv(StringIO(decoded_data))

    def _get(self, url):
        """HTTP GET Request with Retry Logic"""
        url = "{0}/{1}".format(self._base_url, url)
        attempts = 0
        while True:
            try:
                response = self.session.get(url)
                response.raise_for_status()
                return response
            except Exception as e:
                if attempts < self._retry_count:
                    attempts += 1
                    time.sleep(self._retry_delay)
                    continue
                else:
                    raise e

    def get_current_snapshot_id(self, dataset_id):
        """Get ID of the most current snapshot of a dataset"""
        dataset_metadata = self.get_dataset_metadata(dataset_id)
        return dataset_metadata["current_snapshot"]["id"]

    def get_dataset_metadata(self, dataset_id):
        """Get the Dataset Model of this EnigmaReader's dataset
        https://docs.public.enigma.com/resources/dataset/index.html
        """
        url = "datasets/{0}?row_limit=0".format(dataset_id)
        response = self._get(url)
        return response.json()

    def get_snapshot_export(self, snapshot_id):
        """Return raw CSV of a dataset"""
        url = "export/{0}".format(snapshot_id)
        response = self._get(url)
        return response.content



from datetime import datetime

from pandas import DataFrame
import pytest

from pandas_datareader.data import (
    DataReader,
    get_dailysummary_iex,
    get_iex_book,
    get_iex_symbols,
    get_last_iex,
    get_summary_iex,
)
from pandas_datareader.exceptions import UnstableAPIWarning


class TestIEX(object):
    @classmethod
    def setup_class(cls):
        pytest.importorskip("lxml")

    def test_read_iex(self):
        gs = DataReader("GS", "iex-last")
        assert isinstance(gs, DataFrame)

    def test_historical(self):
        df = get_summary_iex(start=datetime(2017, 4, 1), end=datetime(2017, 4, 30))
        assert df.T["averageDailyVolume"].iloc[0] == 137650908.9

    def test_false_ticker(self):
        df = get_last_iex("INVALID TICKER")
        assert df.shape[0] == 0

    @pytest.mark.xfail(
        reason="IEX daily history API is returning 500 as of " "Jan 2018"
    )
    def test_daily(self):
        with pytest.warns(UnstableAPIWarning):
            df = get_dailysummary_iex(
                start=datetime(2017, 5, 5), end=datetime(2017, 5, 6)
            )
            assert df["routedVolume"].iloc[0] == 39974788

    def test_symbols(self):
        df = get_iex_symbols()
        assert "GS" in df.symbol.values

    def test_live_prices(self):
        dftickers = get_iex_symbols()
        tickers = dftickers[:5].symbol.values
        df = get_last_iex(tickers[:5])
        assert df["price"].mean() > 0

    def test_deep(self):
        dob = get_iex_book("GS", service="book")
        if dob:
            assert "GS" in dob
        else:
            pytest.xfail(reason="Can only get Book when market open")



import os

import pandas as pd
import pytest

from pandas_datareader.data import get_sector_performance_av

TEST_API_KEY = os.getenv("ALPHAVANTAGE_API_KEY")
TEST_API_KEY = None if not TEST_API_KEY else TEST_API_KEY

pytestmark = [
    pytest.mark.requires_api_key,
    pytest.mark.alpha_vantage,
    pytest.mark.skipif(TEST_API_KEY is None, reason="ALPHAVANTAGE_API_KEY not set"),
]


class TestAVSector(object):
    @classmethod
    def setup_class(cls):
        pytest.importorskip("lxml")

    @pytest.mark.skipif(TEST_API_KEY is None, reason="ALPHAVANTAGE_API_KEY not set")
    def test_sector(self):
        df = get_sector_performance_av(retry_count=6, pause=20.5)

        cols = pd.Index(["RT", "1D", "5D", "1M", "3M", "YTD", "1Y", "3Y", "5Y", "10Y"])
        assert df.columns.equals(cols)
        assert "Energy" in df.index
        assert "Real Estate" in df.index
